"""Script to simulate an Ecological Momentary Assessment (EMA) study.
Collects all simulated EMA records in files,
as if the data had been gathered in a real experiment,
and runs the analysis.

A simulated experiment may include
one or more Groups of participants randomly drawn from separate Population(s),
with each participant performing several simulated EMA recordings.
Each EMA record includes
(1) a nominal SITUATION, specified in one or more Situation Dimensions,
(2) ordinal ratings of zero, one or more perceptual ATTRIBUTE(s).


*** Usage:
For first demo: Just run this script and look at the generated data files and the results.

For planning an EMA study:
Copy and edit this script and run it for any desired simulation.

Given some anticipated true effect-sizes in the patterns of population characteristics,
the simulation will indicate whether these effects can be reliably demonstrated
in a planned study, given the limited number of participants,
and the limited number of EMA recordings by each participant.


*** Version history:
* Version 1.0.0:
2023-05-17, EmaModel.initialize(..., n_participants_per_comp, ...) defines initial number of GMM components.
            Parameter max_n_comp no longer used.
2023-04-23, Adapted to simplified group-key representation in ema_data.EmaFrame.
            NOTE: changed signatures EmaFrame.setup(), EmaDataSet.load(), and .save().

* Version 0.9.3:
2022-08-17, minor cleanup adapting to minor changes in some package modules.

* Version 0.9.1:
2022-04-04, all result tables generated and saved as Pandas DataFrame instances
2022-03-27, NAP and mean-grades results directly from raw data in ema_data.EmaDataSet
2022-03-21, using Pandas DataFrame format in EmaDataSet, allowing many input file formats

* Version 0.7.1:
2022-01-19, allow set_random_seed for exactly reproducible results

* Version 0.7: minor update to include new calculation and display functions

* Version 0.6:
2021-12-08, tested methods for restrict_attribute, restrict_threshold

* Version 0.5:
2021-11-16, first functional version
2021-11-21, cleaned and tested beta version
"""
# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':
    import numpy as np
    from pathlib import Path
    import logging
    import pickle

    from EmaCalc.ema_simulation import EmaSimPopulation, EmaSimExperiment
    from EmaCalc.ema_simulation import SubjectThurstone, SubjectBradley

    from EmaCalc.ema_data import EmaFrame, EmaDataSet
    from EmaCalc.ema_model import EmaModel
    from EmaCalc.ema_display import EmaDisplaySet
    from EmaCalc import ema_logging, __version__
    from EmaCalc.ema_display_format import harmonize_ylim

    # ------------ (optional) Set simulator random seed for reproducible data generation
    # from EmaCalc.ema_simulation import set_sim_seed
    # set_sim_seed(12345)  # for reproducible results. ONLY FOR TEST OR DEMO

    # ------------------------ Set up working directory and result logging:
    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data'  # or whatever
    result_path = work_path / 'result'  # or whatever

    model_file = 'test_ema_model.pkl'  # saved model file
    experiment_file = 'test_ema_experiment.pkl'  # saved EmaSimExperiment

    ema_logging.setup(save_path=result_path,
                      log_file='run_sim_log.txt')  # to save the log file

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    logger.info(f'*** Running EmaCalc version {__version__} ***')

    # ------ 1: Define Experimental Framework: Situations, Attributes, and Groups

    emf = EmaFrame.setup(situations={'Phase': ('',),  # Test phase(s), only one unlabelled
                                     'HA': ('A', 'B'),  # two Hearing-aid programs
                                     'CoSS': [f'C{i}' for i in range(1, 8)],  # Seven CoSS categories
                                     },
                         phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         groups = {'Age': ['young', 'old'],
                                   'Gender': ['X'], # single category, just as example
                                   }    # -> only TWO distinct groups
                         )
    # NOTE: situations and attributes and group keys will be parts of result file names,
    # so they can include only characters allowed in file names.

    # NOTE: situations and attributes always analyzed as Categorical variables,
    # even if the categories / grades are defined by numeric values.
    # The rank order of attribute grades is defined by the order as listed here,
    # not by, e.g., alphabetical or numerical order of the grade labels.

    # NOTE: if EQUAL grades are used for more than one attribute,
    # the model still assumes separate rating scales for each attribute.
    # However, if grade sequences are IDENTICAL objects for more than one attribute,
    # the model uses the SAME rating scale for those attributes.

    # Response thresholds are always estimated separately for each participant.

    # --------------- Define Population(s) to be simulated:

    # Not necessarily normalized probability-mass arrays
    pop0_situation_prob = np.ones((2, 7))  # with HA = A, B

    # pop0_situation_prob.shape == emf.situation_shape
    # OR == emf.situation_shape[1:] if emf.situation_shape[0] == 1, i.e., only one phase

    # --------- Population mean for Attribute 'Speech', in subject_class scale units

    pop0_speech_mean = np.array([[0., 0., 0., 0., 0., 0., 0.],     # with 'HA' = 'A'
                                 [3., 3., 3., 0., -3., -3., -3]])  # with 'HA' = 'B'
    pop0_speech_mean[1] += 0.5  # add fixed difference between HA (B - A)
    pop0_comfort_mean = pop0_speech_mean[[1, 0], :]  # opposite pattern to 'Speech'

    pop_0 = EmaSimPopulation(emf,
                             situation_prob=pop0_situation_prob,
                             attribute_mean={'Speech': pop0_speech_mean,
                                             'Comfort': pop0_comfort_mean},
                             log_situation_std=0.,  # inter-individual standard deviation of log prob
                             attribute_std=1.0,  # inter-individual standard deviation of attribute locations
                             response_width_mean=1.5,  # response interval width in subject_class units
                             log_response_width_std=0.3,  # inter-individual random threshold variations
                             # subject_class=SubjectBradley,  # default, logistic
                             id='Pop0'
                             )
    # Optionally, define other populations here,
    pop1_situation_prob = pop0_situation_prob
    pop1_speech_mean = np.array([[0., 0., 0., 0., 0., 0., 0.],     # with 'HA' = 'A'
                                 [0., 0., 0., 0., 0., 0., 0.]])  # with 'HA' = 'B'
    pop1_speech_mean[0] -= 1.  # add fixed difference between HA (B - A)
    pop1_speech_mean[1] += 1.  # add fixed difference between HA (B - A)
    pop1_comfort_mean = pop1_speech_mean[[1, 0], :]  # opposite pattern to 'Speech'
    # pop1_comfort_mean = pop0_comfort_mean  # SAME as pop0
    pop_1 = EmaSimPopulation(emf,
                             situation_prob=pop1_situation_prob,
                             attribute_mean={'Speech': pop1_speech_mean,
                                             'Comfort': pop1_comfort_mean},
                             log_situation_std=0.,  # inter-individual standard deviation of log prob
                             attribute_std=0.5,  # inter-individual standard deviation of attribute traits
                             response_width_mean=1.5,  # response interval width in subject_class units
                             log_response_width_std=0.3,  # inter-individual random threshold variations
                             # subject_class=SubjectBradley,  # default
                             id='Pop1'
                             )
    # with SAME EmaFrame, but different situation_prob and/or mean Attribute values

    # ------------------ Generate Experiment with one Group of participants sampled from each Population:
    ema_exp = EmaSimExperiment(emf,
                               groups={('old', 'X'): pop_0.gen_group(n_participants=20),
                                       ('young', 'X'): pop_1.gen_group(n_participants=15)
                                       })
    # NOTE: group keys MUST be a tuple of one group category from each group dimension,
    #   as defined in EmaFrame.setup()

    logger.info(f'ema_exp= {ema_exp}')
    # ema_exp defines true parameter values for Population(s) and all participants in each Group

    # -------- (Optional) Save EmaSimExperiment instance with true parameter values
    data_path.mkdir(parents=True, exist_ok=True)
    with (data_path / experiment_file).open('wb') as f:
        pickle.dump(ema_exp, f)
    logging.info(ema_exp.__class__.__name__ + ' saved to file ' + str(data_path / experiment_file))

    # ------------------------- Generate EMA data for all participants in all Groups:

    ds = ema_exp.gen_dataset(min_ema=30,    # min n of EMA records per participant
                             max_ema=70)    # max n of EMA records per participant
    # = a complete EmaDataSet instance with simulated EMA records for all participants

    # Optionally, save all data in a set of data files, one for each participant
    file_format = 'csv'  # 'xlsx', or other format that Pandas can handle
    ds.save(data_path, fmt=file_format, allow_over_write=True)
    logging.info('Simulated EMA data saved in ' + str(data_path) + f' as {file_format} files')

    # ----------------- (Optional) show mean grades and NAP results for all participants

    mean_grades = ds.attribute_grade_mean(groupby=('HA', 'CoSS'))
    mean_grades.save(result_path / 'Attribute_mean_grades.txt', float_format='%.2f')
    # mean_grades.save(result_path / 'Attribute_mean_grades.csv')  # for input to other analysis
    logger.info(f'Attribute_mean_grades saved in {result_path}')

    # nap = ds.nap_table(sit='HA', nap_cat=['A', 'B'], grouping=('CoSS',))  # grouped results
    nap = ds.nap_table(sit='HA', nap_cat=['A', 'B'], p=0.95)
    # default aggregated across other situation dimensions
    nap.save(result_path / 'NAP.txt')
    # nap.save(result_path / 'NAP.csv')  # for input to other analysis
    logger.info(f'NAP results saved in {result_path}')

    # ------------------------------- Learn Analysis Model from simulated data set:

    # -------- Test re-loading simulated data from saved files (optional):
    logging.info('Simulated EMA data re-loaded from ' + str(data_path) + f' as {file_format} files')
    ds = EmaDataSet.load(emf, data_path, fmt=file_format,
                         participant='file',
                         path_groups=['Age', 'Gender'])

    logger.info(f'Using data ds=\n{ds}')

    # Model ordinal-regression effects of situations on each Attribute:
    # regression_effects = ['HA',     # main linear regression effect only
    #                       'CoSS',   # main linear regression effect only
    #                       # 'Phase',  # if there are several phase categories
    #                       ]
    regression_effects = [('HA', 'CoSS')  # joint effects, main AND interaction
                          # 'Phase',  # if there are several phase categories
                          ]
    # NOTE: A regression_effects element may include any combination of situation dimensions, BUT
    # including ALL interactions -> many model parameters,
    # possibly -> less reliable estimation for each parameter.

    # In this example: ['HA', 'CoSS'] -> 2 + (7 - 1) = 8 regression-effect parameters
    #                ['CoSS', 'HA'] -> 7 + (2 - 1) = 8 regression-effect parameters
    #                [('HA', 'CoSS')] -> 2 * 7 = 14 regression-effect parameters

    emm = EmaModel.initialize(ds,
                              effects=regression_effects,
                              # n_participants_per_comp=5,  # default
                              restrict_attribute=False,  # default
                              restrict_threshold=True,   # default
                              # seed=12345  # ONLY if reproducible results are required
                              )
    # max_n_comp = max number of mixture components in population model
    # restrict_attribute=True -> force attribute mean location at zero
    # restrict_threshold=True -> force one mid-scale response threshold at zero
    # for each respondent and each sample of each attribute

    ll = emm.learn(max_hours=2., max_minutes=0.)
    logger.info(f'*** Data log-likelihood = {ll[-1]:.1f}, indicating model fit to data. ***')
    # *** Recommend re-run model learning with same data set a couple of times, use best-fitting model result.

    emm.prune()  # keep only active mixture components

    # -------- Save learned EmaModel instance (optional):
    with (work_path / model_file).open('wb') as f:
        pickle.dump(emm, f)

    # ------------------------------- generate all result displays:

    emd = EmaDisplaySet.show(emm,
                             situations=['CoSS',  # CoSS probabilities, aggregated across HA
                                         ('CoSS', 'HA'),  # CoSS probabilities, conditional on HA
                                         ('HA', 'CoSS'),  # HA probabilities, conditional on CoSS
                                         ],
                             attributes=[('Speech', 'CoSS'),    # Speech, main effect of CoSS
                                         ('Speech', 'HA'),      # Speech, main effect of HA
                                         ('Speech', ('CoSS', 'HA')),    # joint effect of both
                                         ('Comfort', ('CoSS', 'HA'))],  # joint effect of both
                             grade_counts=['Speech',  # total 'Speech' grade-counts, sum across HA and CoSS
                                           ('Speech', 'HA'),    # 'Speech' grade-counts, separated by HA
                                           ('Comfort', 'HA')],
                             random_individual=True,    # random individual in population
                             population_mean=True,      # population mean
                             participants=False,        # individual results: True -> MANY plots and tables
                             grade_thresholds=True,     # True -> median response thresholds in attribute plots
                             percentiles=[2.5, 25, 50, 75, 97.5],   # in profile plots and tables
                             credibility_limit=0.7,     # minimum credibility in difference tables
                             mpl_params={'figure.max_open_warning': 0,
                                         'axes.labelsize': 'x-large'},  # -> matplotlib.rcParam
                             # mpl_style='my_style_sheet',
                             # ... any other ema_display.FMT or ema_display_format.FMT settings
                             )
    # NOTE: joint (=interaction) effects are correct only if included in EmaModel regression effects

    # ------------------------------- (optionally) edit display elements, if desired
    for g_disp in emd.groups.values():
        harmonize_ylim([g_disp.population_mean.attributes[('Speech', ('CoSS', 'HA'))].plot.ax,
                        g_disp.random_individual.attributes[('Speech', ('CoSS', 'HA'))].plot.ax,
                        g_disp.population_mean.attributes[('Comfort', ('CoSS', 'HA'))].plot.ax,
                        g_disp.random_individual.attributes[('Comfort', ('CoSS', 'HA'))].plot.ax
                        ])
        harmonize_ylim([g_disp.population_mean.situations[('CoSS', 'HA')].plot.ax,
                        g_disp.random_individual.situations[('CoSS', 'HA')].plot.ax,
                        ])

    # ------------------------------- save all result displays
    emd.save(result_path, figure_format='pdf', table_format='txt', float_format='%.2f')
    # (optionally) save in other format(s), too:
    # emd.save(result_path,
    #          table_format='csv',  # for input to other package
    #          float_format='%.4f',  # any other parameters for Pandas table-writer function
    #          # sep='\t'  # -> tab-delimited
    #          )

    logging.info(f'All result displays saved in {result_path}')

    logging.shutdown()



"""This module defines classes and functions to display analysis results
given an ema_model.EmaModel instance,
learned from a set of data from an EMA study.

Results are shown as figures and tables.
Figures can be saved in any file format allowed by Matplotlib, e.g.,
    'pdf', 'png', 'jpg', 'eps'.
Tables can be saved in any file format allowed by Pandas, e.g.,
    'txt', 'tex', 'csv', 'xlsx'.
Thus, both figures and tables can be easily imported into a LaTeX or other word-processing document.

Plot properties may be controlled in detail by keyword arguments to
    method EmaDisplaySet.show(...), either as
1: keyword argument 'mpl_style' with previously defined matplotlib style sheet(s)
2: keyword argument 'mpl_params' as a dict with specific matplotlib parameters
3: setting specific parameters in ema_display.FMT, or ema_display_format.FMT,
    e.g., 'colors' and 'markers'.


*** Main Class:

EmaDisplaySet = a structured container for all display results

Each display element can be accessed and modified by the user, before saving.

The display set can include data for three types of predictive distributions:
*: a random individual in the Population from which a group of respondents was recruited,
    (most relevant for an experiment aiming to predict the success of a new product,
    among individual potential future customers in a population)
*: the mean (=median) in the Population from which a group of test participants were recruited
    (with interpretation most similar to a conventional significance test)
*: each individual respondent in a Group of test participants, with observed EMA data


*** Usage Example:
    See main scripts run_ema and run_sim

Figures and tables are automatically assigned descriptive names,
and saved in a directory tree with names constructed from labels of Groups,
and requested population / participant results.

Thus, after saving, the display files are stored as, e.g.,
result_path / group / 'random_individual' / attributes / ....
result_path / group / 'random_individual' / situations / ....
result_path / group / 'participants' / participant_id / attributes / ....
result_path / group / 'participants' / participant_id / situations / ....
result_path / 'group_effects' / 'population_mean' / attributes / ...  (if more than one group)

*** Version History:
* Version 1.0.0:
2023-04-29, AttributeProfile, SituationProfile work-around for Pandas groupby sort bug, reported,
            to be fixed in Pandas v. 2.0.2 or 2.1 ?
2023-04-22, adapted to simplified group-key representation in ema_data classes

* Version 0.9.6:
2023-04-13, code cleanup: SituationProfile, AttributeProfile, SituationDiff, AttributeDiff
            using Pandas access to model results via new ema_base methods
2023-03-29, include percentile plots and tables in SituationDiff and AttributeDiff objects

* Version 0.9.5:
2023-03-07, include observed and model-estimated grade-count profiles,
            as requested by one reviewer for the Frontiers (2023) paper.
2023-02-27, allow user to set n_samples for population-model display calculations

* Version 0.9.4:
2023-01-22, added default subdirectory names to FMT dict, to allow user control

* Version 0.9.3:
2022-07-27, changed 'subject' -> 'participant',
    'population_individual' -> 'random_individual' in EmaDisplaySet arguments and output.
2022-07-27, allow matplotlib style sheet(s) and individual matplotlib params
    as keyword arguments to EmaDisplaySet.show(...)
2022-07-13, adapted to use new name scenario -> situation

* Version 0.9.2:
2022-06-17, enforce y_min=0 in situation probability percentile plot
2022-06-04, removed FMT['and_head'], not used in ema_display_format.tab_credible_diff

* Version 0.9.1:
2022-04-04, no NAP calculations here, done by ema_data.EmaDataSet instead
2022-04-04, no module-global formatting parameters in ema_display_format, only here
2022-04-04, EmaDisplaySet.save(...) takes file-format arguments
2022-04-04, make result tables as pandas DataFrame objects
2022-03-20, adapted to using Pandas CategoricalDtype instances in EmaFrame

* Version 0.8:
2022-02-15, minor cleanup of scenario profile tables

* Version 0.7:
2021-12-19, include table(s) of individual subject NAP results
2021-12-17, display aggregated Attribute effects weighted by Scenario probabilities

* Version 0.6:
2021-12-08, minor checks against numerical overflow

* Version 0.5.1
2021-11-27, allow NO Attributes in model, check display requests, minor cleanup

* Version 0.5
2021-11-05, copied from PairedCompCalc, modified for EmaCalc
2021-11-09, first functional version
"""
# ***** check if Pandas v >= 2.0.2 solved problem with groupby.quantiles sort bug
# ***** local superclass for pretty-printed repr() ?
# ***** allow several Attributes in single plot ?

import numpy as np
from pathlib import Path
import logging
import string
import pandas as pd

from . import ema_display_format as fmt
from .ema_data import group_dir_str

from samppy import credibility_pd as cred_pd

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

# ---------------------------- Default display parameters
FMT = {'situations': (),    # sequence of situation dimensions or dimension-tuples to display
       'attributes': (),    # sequence of (attribute, situation_effect) to display
                            # situation_effect is a single situation key or tuple of such keys
       'grade_counts': (),  # sequence of (attribute, situation_effect) to display as histogram
       'percentiles': [5., 50., 95.],  # allowing 1, 2, 3, or more percentiles
       'credibility_limit': 0.7,  # min probability of jointly credible differences
       'grade_thresholds': True,  # include median response thresholds in plots
       'random_individual': True,  # show result for random individual in population
       'population_mean': True,  # show result for population mean
       'participants': False,   # show results for each respondent
       'scale_unit': '',  # scale unit for attribute plot axis
       'sit_probability': 'Situation Probability',  # label in figs and tables
       'credibility': 'Credibility',  # heading in difference table
       'population_mean_dir': 'population_mean',  # directory name
       'random_individual_dir': 'random_individual',  # directory name
       'participants_dir': 'participants',  # directory name
       'group_join_str': '_',   # between group dimension and category in path string
       'group_effects_dir': 'group_effects',  # directory name
       'n_samples': 1000,  # number of samples for percentile calculations
       }

DEFAULT_FIGURE_FORMAT = 'pdf'
DEFAULT_TABLE_FORMAT = 'txt'


def set_format_param(**kwargs):
    """Set / modify format parameters
    Called before any displays are generated.
    :param kwargs: dict with any formatting variables
    :return: None
    """
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    FMT['percentiles'].sort()  # ensure increasing values
    if len(other_fmt) > 0:
        # may include 'mpl_style' and 'mpl_params', and more...
        fmt.set_format_param(**other_fmt)


# ------------------------------------------------ Elementary Display Class:
class Profile:
    """Container for ONE requested profile display
    for ONE tuple of one or more situation dimensions,
    OR for ONE (attribute, situation_effect) request.
    """
    def __init__(self, plot=None, tab=None, diff=None,
                 path=None):
        """
        :param plot: (optional) ema_display_format.ResultPlot instance with profile plot
        :param tab: (optional) ema_display_format.ResultTable instance with same results tabulated.
        :param diff: (optional) ema_display_format.DiffTable instance with credible differences.
        :param path: (optional) Path to directory containing saved files from this Profile,
            assigned by save() method.
            File names are defined by sub-objects.
        """
        self.plot = plot
        self.tab = tab
        self.diff = diff
        self.path = path

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    def save(self, path,
             figure_format=None,
             table_format=None,
             **kwargs):
        """Save plot and table displays to files.
        May be called repeatedly with different file formats and kwargs.
        :param path: path to directory for saving files
        :param figure_format: (optional) single figure-file format string
        :param table_format: (optional) single table-file format string
        :param kwargs: (optional) additional parameters passed to ResultTable.save() method.
            NOTE: NO extra kwargs allowed for plot.save() method!
        :return: None
        """
        if figure_format is None and table_format is None:
            figure_format = DEFAULT_FIGURE_FORMAT
            table_format = DEFAULT_TABLE_FORMAT
        # One of them may be None -> only the other type of data are saved.
        path.mkdir(parents=True, exist_ok=True)
        self.path = path
        if figure_format is not None:
            if self.plot is not None:
                self.plot.save(path, figure_format=figure_format)
        if table_format is not None:
            if self.tab is not None:
                self.tab.save(path, table_format=table_format, **kwargs)
            if self.diff is not None:
                self.diff.save(path, table_format=table_format, **kwargs)


# ---------------------------------------------------------- Top Display Class:
class EmaDisplaySet:
    """Root container for all displays
    of selected predictive situation and attribute results
    from one ema_model.EmaModel instance learned from one ema_data.EmaDataSet instance.

    All display elements can be saved as files in a selected directory three.
    The complete instance can also be serialized and dumped to a pickle file,
    then re-loaded, edited, and re-saved, if user needs to modify some display element(s).
    """
    def __init__(self, group_head, groups, group_effects=None):
        """
        :param group_head: tuple group dimension labels, copied form EmaFrame instance
        :param groups: dict with (group_id, GroupDisplaySet) elements
        :param group_effects: (optional) single GroupEffectSet instance,
            showing jointly credible differences between groups,
            IFF there is more than one group
        """
        self.group_head = group_head  # copy from EmaModel.emf.group_head()
        self.groups = groups
        self.group_effects = group_effects

    def __repr__(self):  # *** general superclass for repr?
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    def save(self, dir_top, **kwargs):
        """Save all displays in a directory tree
        :param dir_top: Path or string with top directory for all displays
        :param kwargs: (optional) dict with any additional format parameters, e.g.,
            figure_format, table_format, and any Pandas file-writer parameters.
        :return: None
        """
        dir_top = Path(dir_top)
        for (g, g_display) in self.groups.items():
            # g = _dir_name(self.group_head, g, sep='/')
            g = group_dir_str(self.group_head, g, sep=FMT['group_join_str'])
            if len(g) == 0 or all(s in string.whitespace for s in g):
                g_display.save(dir_top, **kwargs)
            else:
                g_display.save(dir_top / g, **kwargs)
        if self.group_effects is not None:
            self.group_effects.save(dir_top / FMT['group_effects_dir'], **kwargs)

    @classmethod
    def show(cls, emm,
             situations=None,
             attributes=None,
             grade_counts=None,
             **kwargs):
        """Create requested displays for results from an EMA study,
        and store all display elements in a single structured object.
        :param emm: an ema_model.EmaModel instance, where
            emm.groups[group] is an ema_model.EmaGroupModel instance,
            emm.groups[group][participant_id] is an ema_model.EmaRespondentModel instance
        :param situations: (optional) list with selected situation dimensions to be displayed.
            situations[i] = a selected key in emm.emf.situation_dtypes, or a tuple of such keys.
        :param attributes: (optional) list with selected attribute displays to be displayed.
            attributes[i] = a tuple (attribute_key, sit_effect), where
                attribute_key is one key in emm.emf.attribute_grades,
                sit_effect is a situation dimension in emm.emf.situation_dtypes, or a tuple of such keys.
                A single key will yield the main effect of the named situation dimension.
                An effect tuple will also show interaction effects between situation dimensions,
                IFF the regression model has been set up to estimate such interaction effects.
        :param grade_counts: (optional) list with selected attribute grade counts to be displayed,
                given as single attribute key, or tuple (attribute_key, situation) .
        :param: kwargs: (optional) dict with any other display formatting parameters
            for ema_display.FMT and / or ema_display_format.FMT and / or matplotlib
        :return: a cls instance filled with display objects
        """
        # get default scale_unit from emm, if not in kwargs:
        if 'scale_unit' not in kwargs:
            kwargs['scale_unit'] = emm.base.rv_class.unit_label
        if situations is None:
            situations = [*emm.base.emf.situation_dtypes.keys()]
            # default showing only main effects for each situation dimension
        situations = [(sit,) if isinstance(sit, str) else sit
                      for sit in situations]  # ensure tuple elements
        # *** check that requested situations exist in model:
        missing_sit = [sit_tuple for sit_tuple in situations
                       if any(sit_ij not in emm.base.emf.situation_dtypes.keys()
                              for sit_ij in sit_tuple)]
        for sit in missing_sit:
            logger.warning(f'Situation dimension in {sit} unknown in the learned model.')
            situations.remove(sit)
        if attributes is None:
            attributes = []
        attributes = [(a, sit) if type(sit) is tuple else (a, (sit,))
                      for (a, sit) in attributes]
        # *** check that requested attribute effects exist in model
        missing_attr = [a_sit for a_sit in attributes
                        if a_sit[0] not in emm.base.emf.attribute_dtypes.keys()]
        for a_sit in missing_attr:
            logger.warning(f'Attribute effect {a_sit} unknown in the learned model.')
            attributes.remove(a_sit)
        if grade_counts is not None:  # ensure tuple
            grade_counts = [(gc,) if isinstance(gc, str) else gc
                            for gc in grade_counts]
        set_format_param(situations=situations,
                         attributes=attributes,
                         grade_counts=grade_counts,
                         **kwargs)
        # display separate results for each group
        group_head = emm.base.emf.group_head()
        groups = {g: GroupDisplaySet.display(emm_g)
                  for (g, emm_g) in emm.groups.items()}

        if len(groups) > 1:
            group_effects = GroupEffectSet.display(emm)
        else:
            group_effects = None
        logger.info(fig_comments())
        logger.info(table_comments())
        return cls(group_head, groups, group_effects)


class GroupDisplaySet:
    """Container for all quality displays related to ONE study group:
    Predictive results for the population from which the participants were recruited,
    and for each individual participant, if requested by user.
    """
    def __init__(self,
                 population_mean=None,
                 random_individual=None,
                 participants=None,
                 counts=None):
        """
        :param population_mean: EmaDisplay instance, for population mean,
        :param random_individual: EmaDisplay instance, for random individual,
        :param participants: dict with elements (s_id: EmaDisplay instance) for participants in one group
            with situation and attribute results.
        :param counts: CountDisplay instance, for observed and model-estimated count profiles
        """
        self.population_mean = population_mean
        self.random_individual = random_individual
        self.participants = participants
        self.counts = counts

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    def save(self, path, **kwargs):
        """Save all stored display objects in their corresponding subdirectories
        """
        if self.population_mean is not None:
            self.population_mean.save(path / FMT['population_mean_dir'], **kwargs)
        if self.random_individual is not None:
            self.random_individual.save(path / FMT['random_individual_dir'], **kwargs)
        if self.participants is not None:
            for (s, s_disp) in self.participants.items():
                s_disp.save(path / FMT['participants_dir'] / str(s), **kwargs)
        if self.counts is not None:
            self.counts.save(path / 'counts', **kwargs)  # *****************

    @classmethod
    def display(cls, emm_g):
        """Generate all displays for ONE group
        :param emm_g: dict with elements (a_label: ema_group.EmaGroupModel)
        :return: cls instance with all displays for this group
        """
        pop_ind = None
        pop_mean = None
        participants = None
        counts = None
        if FMT['random_individual']:
            pop_ind = EmaDisplay.display(emm_g.predictive_population_ind())
        if FMT['population_mean']:
            pop_mean = EmaDisplay.display(emm_g.predictive_population_mean())
        if FMT['participants']:
            # logger.debug('Displaying participants:')
            participants = {s: EmaDisplay.display(s_model)
                            for (s, s_model) in emm_g.participants.items()}
        if FMT['grade_counts'] is not None and len(FMT['grade_counts']) > 0:
            counts = CountDisplay.show(emm_g)
        return cls(population_mean=pop_mean,
                   random_individual=pop_ind,
                   participants=participants,
                   counts=counts)


# ------------------------------------- Secondary-level displays
class EmaDisplay:
    """Container for all situation and attribute displays of model-estimated parameters
    for ONE (Sub-)Population, either random-individual or mean,
    OR for ONE participant.
    """
    def __init__(self, situations, attributes):
        """
        :param situations: dict with (situation_tuple, profile), where
            profile is a Profile instance for the selected situation_tuple
        :param attributes: dict with (attr_effect, profile), where
            profile is a Profile instance for the selected attr_effect
        """
        self.situations = situations
        self.attributes = attributes

    def save(self, path, **kwargs):
        """Save all stored display objects in specified (sub-)tree
        :param path: path to directory for saving
        :return: None
        """
        if len(self.situations) > 0:
            for (d, sit_display) in self.situations.items():
                sit_display.save(path / 'situations', **kwargs)
        if len(self.attributes) > 0:
            for (d, a_display) in self.attributes.items():
                a_display.save(path / 'attributes', **kwargs)

    @classmethod
    def display(cls, m_xi):
        """
        :param m_xi: probability model for ema_model parameter vector xi
            either for random individual, population mean, or individual participant
        :return: a cls instance
        """
        xi = m_xi.rvs(size=FMT['n_samples'])
        logger.debug(f'Using {len(xi)} samples')
        situations = {sit: SituationProfile.display(xi, m_xi, sit)
                      for sit in FMT['situations']}
        attributes = {a_effect: AttributeProfile.display(xi, m_xi, a_effect)
                      for a_effect in FMT['attributes']}
        return cls(situations, attributes)


class CountDisplay:
    """Container for attribute displays of grade-count distributions,
    including both observed and model-estimated EMA-counts,
    for ONE (Sub-)Population, (OR for ONE participant ? ********).
    """
    def __init__(self, attributes):
        """
        :param attributes: dict with (attr_effect, count-profile), where
            profile is a Profile instance for the selected attr_effect
        """
        self.attributes = attributes

    def save(self, path, **kwargs):
        """Save all stored display objects in specified (sub-)tree
        :param path: path to directory for saving
        :return: None
        """
        if len(self.attributes) > 0:
            for a_display in self.attributes.values():
                a_display.save(path, **kwargs)

    @classmethod
    def show(cls, g_model):
        """
        :param g_model: probability model for ema_model parameter vector xi
            either for random individual, population mean, or individual participant
        :return: a cls instance
        """
        def show_histogram(a_effect):
            if len(a_effect) == 1:
                (a, sit_case) = (a_effect[0], None)
            else:
                (a, sit_case) = a_effect
            df_obs = g_model.attribute_grade_count(a, groupby=sit_case)
            df_mod = g_model.rvs_grade_count(a, groupby=sit_case)
            q = np.array(FMT['percentiles']) / 100
            if sit_case is None:
                # df_obs is a Series object, make it a DataFrame with one row
                df_obs = df_obs.to_frame().T
                df_q = df_mod.quantile(q, numeric_only=True)
                fill_value = df_obs.index.values[0]
                i = ((fill_value, q) for q in df_q.index.values)
                df_q = df_q.set_index(keys=i)
            else:
                df_q = df_mod.groupby(sit_case,
                                      sort=False, group_keys=True).quantile(q,
                                                                            numeric_only=True)
            return Profile(plot=fmt.fig_category_barplot(df=df_obs, df_q=df_q,
                                                         x_label=a + ': Ordinal Grades',
                                                         y_label='EMA counts',
                                                         file_label=a)
                           )
        # ------------------------------------------

        attributes = {a_effect: show_histogram(a_effect)
                      for a_effect in FMT['grade_counts']}
        return cls(attributes)


# ----------------------------------------------------------------------
class SituationProfile(Profile):
    """Container for all displays of situation probability profiles
    in ONE (Sub-)Population, OR for ONE respondent.
    NOTE: situation profiles are displayed as
    CONDITIONAL probability of categories in ONE situation dimension,
    GIVEN categories in other dimension(s), if requested.
    """
    @classmethod
    def display(cls, xi, m_xi, sit_keys):
        """Generate a probability-profile display for selected distribution and factor
        :param xi: 2D array of parameter-vector samples drawn from m_xi
        :param m_xi: a population or individual model instance
        :param sit_keys: tuple of one or more key(s) selected from emf.situation_dtypes.keys()
        :return: single cls instance showing CONDITIONAL probabilities
            for sit_keys[0], GIVEN each combination (j1,..., jD) for sit_keys[1], ...
        """
        u_ds = m_xi.base.situation_prob_df(xi, groupby=sit_keys)
        # = pd.Series object with MultiIndex axes [sample, *sit_keys]
        # quantiles = np.array(FMT['percentiles']) / 100.
        # q_ds = u_ds.groupby(level=list(sit_keys),
        #                     sort=False, group_keys=True).quantile(quantiles,
        #                                                           numeric_only=True)
        # # = pd.Series object with MultiIndex axes [*sit_keys, quantiles]
        q_ds = _series_quantile_fix(u_ds, sit_keys)  # *** TEMP fix for Pandas bug
        tab_perc = fmt.tab_percentiles(q_ds)
        fig_perc = fmt.fig_percentiles(tab_perc, y_label=FMT['sit_probability'], file_label='', y_min=0.)
        # ---------------------------------------- sit_keys differences
        # NOTE: Comparing CONDITIONAL probabilities of categories in FIRST sit_keys dimension,
        # GIVEN categories in other dimensions.
        d_pd = cred_pd.cred_diff(u_ds, diff_axis=sit_keys[0], case_axis=sit_keys[1:], p_lim=FMT['credibility_limit'])
        tab_diff = fmt.tab_credible_diff(d_pd, diff_head=sit_keys[0:1], cred_head=FMT['credibility'],
                                         case_head=sit_keys[1:], y_label=FMT['sit_probability'])
        # ---------------------------------------------------------------------
        return cls(plot=fig_perc, tab=tab_perc, diff=tab_diff)


class AttributeProfile(Profile):
    """Container for displays of ONE attribute value effect of situation(s),
    in ONE (Sub-)Population, OR for ONE respondent.

    NOTE: Latent-variable results are displayed for each Attribute,
    GIVEN Situation categories in requested Situation dimensions,
    averaged across all OTHER Situation dimensions,
    weighted by Situation probabilities in those dimensions.
    """
    # **** allow several attributes in one display ? *****
    @classmethod
    def display(cls, xi, m_xi, a_effect):
        """Create displays for a single attribute and requested situation effects
        :param xi: 2D array of parameter-vector samples drawn from m_xi
        :param m_xi: a population or individual model instance
        :param a_effect: tuple(attribute_key, sit_keys)
        :return: single cls instance with all displays
        """
        (a, sit_keys) = a_effect
        # --------------------------------------- thresholds, optional:
        if FMT['grade_thresholds']:
            tau = np.median(m_xi.base.attribute_tau(xi, a), axis=0)
            # tau[l] = l-th median rating threshold for attribute a, SAME for all situations
        else:
            tau = None
        # --------------------------------------- percentile table:
        theta_ds =  m_xi.base.attribute_theta_df(xi, a, groupby=sit_keys)
        # = pd.Series with MultiIndex [samples, *sit_keys]
        # ***** with UN-sorted sit_keys
        # quantiles = np.array(FMT['percentiles']) / 100.
        # if len(sit_keys) > 1:
        #     theta_q = theta_ds.groupby(level=sit_keys,
        #                                sort=False, group_keys=True).quantile(quantiles)
        # # *** with SORTED index.levels for sit_key if only one dimension. Pandas v 2.0.0, 2.0.1
        # # *** UN-sorted index.levels, if sit_keys is more than one level ***
        # else:  # *** work-around for Pandas bugg ***
        #     theta_q = pd.concat({s: q.quantile(quantiles)
        #                          for (s, q) in theta_ds.groupby(level=sit_keys, sort=False)},
        #                         names=list(sit_keys))
        theta_q = _series_quantile_fix(theta_ds, sit_keys)  # *** TEMP fix for Pandas bug
        tab_perc = fmt.tab_percentiles(theta_q, file_label=a)
        fig_perc = fmt.fig_percentiles(tab_perc, y_label=a + ' (' + str(FMT['scale_unit']) + ')',
                                       file_label=a,
                                       cat_limits=tau)
        # ---------------------------------------- attr differences
        # NOTE: comparing all situation-categories, in all requested sit_keys dimensions
        d_pd = cred_pd.cred_diff(theta_ds, diff_axis=sit_keys, p_lim=FMT['credibility_limit'])
        tab_diff = fmt.tab_credible_diff(d_pd, diff_head=sit_keys, cred_head=FMT['credibility'],
                                         y_label=a,
                                         file_label=a)
        return cls(plot=fig_perc, tab=tab_perc, diff=tab_diff)


# --------------------------------- classes for differences between groups
class GroupEffectSet:
    """Container for displays of differences between populations,
    as represented by participants in separate groups
    """
    def __init__(self, population_mean=None, random_individual=None):
        """
        :param population_mean: EmaGroupDiff instance
        :param random_individual: EmaGroupDiff instance
            both with results separated by groups
        """
        self.population_mean = population_mean
        self.random_individual = random_individual

    def save(self, path, **kwargs):
        """Save all stored display objects in their corresponding sub-trees
        """
        if self.population_mean is not None:
            self.population_mean.save(path / FMT['population_mean_dir'], **kwargs)
        if self.random_individual is not None:
            self.random_individual.save(path / FMT['random_individual_dir'], **kwargs)

    @classmethod
    def display(cls, emm):
        """Generate all displays for ONE group
        :param emm: ema_model.EmaModel instance with several groups
        :return: cls instance with all displays of differences
            between predictive distributions for
            population random individual, AND/OR
            population mean
        """
        pop_ind = None
        pop_mean = None
        if FMT['random_individual']:
            pop_ind = EmaGroupDiff.display(emm,
                                           {g: emm_g.predictive_population_ind()
                                            for (g, emm_g) in emm.groups.items()})
        if FMT['population_mean']:
            pop_mean = EmaGroupDiff.display(emm,
                                            {g: emm_g.predictive_population_mean()
                                             for (g, emm_g) in emm.groups.items()})
        return cls(population_mean=pop_mean,
                   random_individual=pop_ind)


class EmaGroupDiff:
    """Container for displays of differences between (Sub-)Populations
    represented by separate ema_model.EmaGroupModel instances.
    """
    def __init__(self, situations, attributes):
        """
        :param situations: dict with (situation_tuple, SituationDiff instance)
        :param attributes: dict with (attr_effect, AttributeDiff instance), where
        """
        self.situations = situations
        self.attributes = attributes

    def save(self, path, **kwargs):
        """Save all stored display objects in specified (sub-)tree
        """
        if len(self.situations) > 0:
            for (d, sit_display) in self.situations.items():
                sit_display.save(path / 'situations', **kwargs)
        if len(self.attributes) > 0:
            for (d, a_display) in self.attributes.items():
                a_display.save(path / 'attributes', **kwargs)

    @classmethod
    def display(cls, emm, groups):
        """
        :param emm: ema_model.EmaModel instance with several groups
        :param groups: dict with elements (g_id, g_model), where
            g_id is a tuple of one or more tuple(group_factor, factor_category),
            g_model is a predictive ema_model.PredictivePopulationModel instance
            i.e. NOT emm.groups
        :return: single cls instance
        """
        xi = [g_model.rvs(size=FMT['n_samples'])
              for g_model in groups.values()]
        # xi[g][s, :] = s-th sample of parameter vector for g-th group
        situations = {sc: SituationDiff.display(xi, emm, sc)
                      for sc in FMT['situations']}
        attributes = {a_effect: AttributeDiff.display(xi, emm, a_effect)
                      for a_effect in FMT['attributes']}
        return cls(situations, attributes)


class SituationDiff(Profile):
    @classmethod
    def display(cls, xi, emm, sit_keys):
        """Generate a probability-profile display for selected distribution and factor
        :param xi: list of 2D arrays of parameter-vector samples
            len(xi) == len(emm.groups)
        :param emm: ema_model.EmaModel object
        :param sit_keys: tuple of one or more key(s) selected from emf.situation_dtypes.keys()
        :return: single cls instance
        """
        # --------------------------------- situation prob. vs (sit_keys, groups):
        group_head = emm.base.emf.group_head()  # ********************
        group_cat = list(emm.groups.keys())
        u_groups = {g: emm.base.situation_prob_df(xi_g, groupby=sit_keys)
                    for (g, xi_g) in zip(group_cat, xi)}
        # = dict of pd.Series objects, each with MultiIndex [sample, *sit_keys]
        u_ds = pd.concat(u_groups, axis=0, names=list(group_head))
        quantiles = np.array(FMT['percentiles']) / 100.
        u_q = u_ds.groupby(level=list(sit_keys) + list(group_head),
                           sort=False, group_keys=True).quantile(quantiles,
                                                                 numeric_only=True)
        # = pd.Series with MultiIndex [*sit_keys, quantiles]
        tab_perc = fmt.tab_percentiles(u_q)
        fig_perc = fmt.fig_percentiles(tab_perc, y_label=FMT['sit_probability'], file_label='', y_min=0.)
        # ---------------------------------------- group differences by sit_keys
        # NOTE: Comparing CONDITIONAL probabilities of categories in FIRST sit_keys dimension,
        # GIVEN categories in other dimensions.
        d_pd = cred_pd.cred_group_diff(u_groups, group_axis=group_head,
                                       case_axis=sit_keys,
                                       p_lim=FMT['credibility_limit'])
        tab_diff = fmt.tab_credible_diff(d_pd, diff_head=group_head, cred_head=FMT['credibility'], case_head=sit_keys,
                                         y_label=FMT['sit_probability'])
        return cls(plot=fig_perc, tab=tab_perc, diff=tab_diff)


class AttributeDiff(Profile):
    """Container for all displays of group differences in ONE Attribute-by-Situation effect
    """
    @classmethod
    def display(cls, xi, emm, a_effect):
        """Create displays for a single attribute and requested situation effects
        :param xi: list of 2D arrays of parameter-vector samples
            len(xi) == len(emm.groups)
        :param a_effect: tuple (attr_key, sit_keys), where
            sit_keys is a tuple of one or more key(s) selected from emf.situation_dtypes.keys()
        :param emm: ema_model.EmaModel object
        :return: single cls instance
        """
        (a, sit_keys) = a_effect
        if FMT['grade_thresholds']:
            tau = np.array([emm.base.attribute_tau(xi_g, a)
                     for xi_g in xi])
            # tau[g, s, :] = s-th sample of threshold array for g-th group
            tau = tau.reshape((-1, tau.shape[-1]))
            # tau[gs, :] = gs-th sample across all groups
            tau = np.median(tau, axis=0)
            # tau[l] = l-th median rating threshold for attribute a
        else:
            tau = None
        # --------------------------------------- percentile table
        group_head = emm.base.emf.group_head()
        group_cat = list(emm.groups.keys())
        theta_groups = {g: emm.base.attribute_theta_df(xi_g, a, groupby=sit_keys)
                        for (g, xi_g) in zip(group_cat, xi)}
        theta_ds = pd.concat(theta_groups, axis=0, names=list(group_head))
        quantiles = np.array(FMT['percentiles']) / 100.
        theta_q = theta_ds.groupby(level=list(sit_keys) + list(group_head),
                                   sort=False, group_keys=True).quantile(quantiles,
                                                                         numeric_only=True)
        tab_perc = fmt.tab_percentiles(theta_q, file_label=a)
        fig_perc = fmt.fig_percentiles(tab_perc, y_label=a + ' (' + str(FMT['scale_unit']) + ')', file_label=a,
                                       cat_limits=tau)
        # ---------------------------------------- attribute differences between groups
        # NOTE: comparing all situation-categories, in all requested sc dimensions
        d_pd = cred_pd.cred_group_diff(theta_groups,
                                       group_axis=group_head,
                                       case_axis=sit_keys,
                                       p_lim=FMT['credibility_limit'])
        tab_diff = fmt.tab_credible_diff(d_pd, diff_head=group_head,
                                         cred_head=FMT['credibility'],
                                         case_head=sit_keys,
                                         y_label=a, file_label=a)
        return cls(plot=fig_perc, tab=tab_perc, diff=tab_diff)


# ---------------------------------- Help functions:

def _series_quantile_fix(ds, sit_keys):  # *** Not needed after bug fix in Pandas >= 2.0.2 ?
    """Work-around for possible Pandas bug in groupby with sort=False
    :param ds: pandas.Series instance with Multi-index [_sample, *sit_keys]
    :param sit_keys: tuple with one or more index levels in ds
    :return: series with Multi-Index levels [*sit_keys, quantiles]
    """
    quantiles = np.array(FMT['percentiles']) / 100.
    if len(sit_keys) > 1:
        q = ds.groupby(level=sit_keys,
                                   sort=False, group_keys=True).quantile(quantiles)
    # *** -> SORTED index.levels for sit_key if only one dimension. Pandas v 2.0.0, 2.0.1
    # *** -> OK UN-sorted index.levels, if sit_keys is more than one level ***
    else:  # *** work-around for Pandas bugg ***
        q = pd.concat({s: q.quantile(quantiles)
                       for (s, q) in ds.groupby(level=sit_keys, sort=False)},
                      names=list(sit_keys))
    return q


def fig_comments():
    """Generate figure explanations.
    :return: comment string
    """
    p_min = np.amin(FMT['percentiles'])
    p_max = np.amax(FMT['percentiles'])
    c = f"""Figure Explanations:
    
Figure files with names like
someSituationName_xxx.pdf, someAttributeName_xxx.pdf or xxx.jpg, or similar,
display user-requested percentiles (markers) and credible intervals (vertical bars) 
The vertical bars show the range between {p_min:.1f}- and {p_max:.1f}- percentiles.

Median ordinal grade thresholds for perceptual Attributes, if requested,
are indicated by thin lines extending horizontally across the graph.

The displayed ranges include all uncertainty
caused both by real inter-individual perceptual differences
and by the limited number of responses by each individual.
"""
    if FMT['grade_counts'] is not None and len(FMT['grade_counts']) > 0:
        c += f"""
Attribute grade-counts are shown as bar plots for RECORDED EMA data.
A vertical line on top of each bar shows the range 
between {p_min:.1f}- and {p_max:.1f}- percentiles of corresponding MODEL-PREDICTED grade-counts.
"""
    return c


def table_comments():
    c = """Table Explanations:

*** Tables of Percentiles:
Files with names like someSituationName_xxx.tex, someAttributeName_xxx.tex or *.txt
show numerical versions of the information in corresponding percentile plots.
Percentiles, credible ranges, and marginal probabilities for negative and positive values are shown.

*** Tables of Credible Differences:
Files with names like someAttribute-diff_xxx.tex or *.txt 
show a list of Attribute (in Situation) pairs
which are ALL JOINTLY credibly different
with the tabulated credibility.
The credibility value in each row is the JOINT probability
for the pairs in the same row and all rows above it.
Thus, the joint probability values already account for multiple comparisons,
so no further adjustments are needed.
"""
    return c



"""This module defines classes for a Bayesian probabilistic model of EMA data,
for ONE group of respondents representing ONE population.

*** Class Overview:

EmaGroupModel: Container for individual response-probability models,
    implemented by an ema_respondent.EmaRespondentModel instance for each participant
    in ONE group of test participants, assumed recruited from ONE population.
    Also contains a PopulationModel instance representing the population
    from which the participants were recruited.

PopulationModel: Defines a Gaussian Mixture Model (GMM)
    for the parameter distribution in ONE population,
    from which participants in ONE group were recruited.
    Part of EmaGroupModel, used as prior for all participants in the group.

PredictivePopulationModel: marginal distribution of parameters
    in ONE population represented by ONE EmaGroupModel.
    Specifies a mixture of Student-t distributions,
    used for result displays.

*** Version History:
* Version 1.0.0:
2023-05-17, EmaGroupModel.initialize(...,n_participants_per_comp,...) to set initial number of GMM components
2023-05-15, logger.info participant GMM weights after pruning

* Version 0.9.3:
2022-07-27, changed 'subject' -> 'participant' or 'respondent'

* Version 0.8.3:
2022-03-07, changed class name GroupModel -> EmaGroupModel
2022-03-07, changed class name ProfileMixtureModel -> PredictivePopulationModel
2022-03-07, separate class for GMM: EmaGroupModel.pop_gmm = a PopulationModel instance

* Version 0.8.2:
2022-03-03, allow multi-processing Pool in EmaGroupModel.adapt, parallel across subjects
2022-03-02, test prepare to allow multiprocesing Pool.imap
            allow multi-processing Pool in EmaGroupModel.adapt, parallel across subjects

* Version 0.8.1
2022-02-26, complete separate GMM for each group, GMM components -> EmaGroupModel property comp
"""
# *** test multiprocessing start method fork? Does not work in windows ! *********
# *** Future: try using shared_memory or Ray to reduce multiprocessing overhead?
import copy
import logging
import os
from multiprocessing.pool import Pool

import numpy as np
from scipy.special import logsumexp, softmax
# import pandas as pd

from EmaCalc.dirichlet import DirichletVector
from EmaCalc.dirichlet import JEFFREYS_CONC
from EmaCalc.ema_respondent import EmaRespondentModel

MixtureWeight = DirichletVector

# -------------------------------------------------------------------
__ModelVersion__ = "2022-03-08"

PRIOR_MIXTURE_CONC = JEFFREYS_CONC
# = prior conc for population mixture weights.

N_SAMPLES = 1000
# = default number of parameter samples in PredictivePopulationModel.rvs() calculation

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

usePool = True
# usePool = False # for TEST
# multi-processing speeds up the learning, but the Pool causes some overhead,
# so the time reduction is only about a factor 2 with 4 CPU cores.


# -------------------------------------------------------------------
class EmaGroupModel:
    """Container for EmaRespondentModel instances for all test participants
    in ONE group of respondents,
    and a PopulationModel instance defining
    a Gaussian Mixture Model (GMM) for the parameter distribution in
    the population from which participants were recruited.
    The GMM is prior for the parameter distribution in all EmaRespondentModel instances.
    """
    def __init__(self, base, participants, pop_gmm):
        """
        :param base: single common EmaParamBase object, used by all model parts
        :param participants: dict with (participant_id, EmaRespondentModel) elements
        :param pop_gmm: a PopulationModel instance representing a Gaussian Mixture Model (GMM)
            for the parameter distribution in the population from which participants were recruited.
        """
        self.base = base
        self.participants = participants
        # *** might be just a list, not dict, as each participant has ID as property ***
        self.pop_gmm = pop_gmm
        for s in self.participants.values():
            s.prior = self.pop_gmm

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                f'\n\tparticipants= {len(self.participants)} individuals,' +
                f'\n\tpop_gmm={repr(self.pop_gmm)})'
                )

    @classmethod
    def initialize(cls, n_participants_per_comp, base, group_data, seed_seq, rng):
        """Crude initial group model given group EMA data
        :param n_participants_per_comp: scalar integer = desired number of participants per GMM component
        :param base: single common EmaParamBase object, used by all model parts
        :param group_data: a dict with elements (s_id, s_ema), where
            s_id = a participant key,
            s_ema = a Pandas.DataFrame with EMA records from an ema_data.EmaDataSet object.
        :param seed_seq: SeedSequence object to spawn children for participants of this group
        :param rng: random Generator for this group, used by self.pop_gmm
        :return: a cls instance crudely initialized
        """
        participant_rng_list = [np.random.default_rng(s)
                                for s in seed_seq.spawn(len(group_data))]
        s_models = {s_id: EmaRespondentModel.initialize(base, s_ema, s_rng, id=s_id)
                    for ((s_id, s_ema), s_rng) in zip(group_data.items(),
                                                      participant_rng_list)}
        n_participants = len(s_models)
        # if max_n_comp is None:
        #     max_n_comp = n_participants // 2
        # else:
        #     max_n_comp = min(n_participants // 2, max_n_comp)
        n_comp = max(1, n_participants // n_participants_per_comp)
        pop_gmm = PopulationModel.initialize(n_comp, s_models, base, rng)
        return cls(base, s_models, pop_gmm)

    def adapt(self, g_name):
        """One VI adaptation step for all model parameters
        :param g_name: group id label for logger output
        :return: ll = scalar VI lower bound to data log-likelihood,
            incl. negative contributions for parameter KLdiv re priors

        NOTE: All contributions to VI log-likelihood
        are calculated AFTER all updates of factorized parameter distributions
        because all must use the SAME parameter distribution
        """
        ll_pop_gmm = self.pop_gmm.adapt(g_name, self.participants)
        # = -KLdiv{ q(mix_weight) || prior.mix_weight}
        #   sum_m -KLdiv(current q(mu_m, Lambda_m) || prior(mu_m, Lambda_m))

        # *** allow multiprocessing across participants in parallel
        if usePool:
            n_pool = _pool_size(len(self.participants))
            ch_size = len(self.participants) // n_pool
            # ****** check time ! ********************
            if n_pool * ch_size < len(self.participants):
                ch_size += 1
            logger.debug(f'adapt Pool: n_pool= {n_pool}, ch_size= {ch_size}')
            with Pool(n_pool) as p:
                self.participants = dict(p.imap(_adapt_respondent, self.participants.items(),
                                                chunksize=ch_size))
                # ******* send self.pop_gmm as argument to _adapt_participant ??? ******
                # if so, MUST delete the copy in sub-process! Maybe pickling takes longer time? ***
                for s in self.participants.values():
                    s.prior = self.pop_gmm
                    # MUST restore link; sub-processed object links to a copy of self.pop_gmm
        else:
            self.participants = dict(map(_adapt_respondent, self.participants.items()))
        ll_participants = sum(s_model.ll for s_model in self.participants.values())
        # = sum <ln p(data | xi)> + <ln prior p(xi | comp)> - <ln q(xi)>
        #           - KLdiv(q(zeta | xi) || p(zeta | self.mix_weight)
        #  All ll contributions now calculated using the current updated q() distributions
        logger.debug(f'Group {repr(g_name)}: '
                     + f'\n\tll_pop_gmm = {ll_pop_gmm:.3f}'
                     + f'\n\tparticipant sum ll_xi= {ll_participants:.3f}')
        return ll_pop_gmm + ll_participants

    def prune(self, g_name, min_weight=JEFFREYS_CONC):
        """Prune model to keep only active mixture components
        :param g_name: group label for logger output
        :param min_weight: scalar, smallest accepted value for sum individual weight
        :return: None
        Result: all model components pruned consistently
        """
        w_sum = np.sum([np.mean(self.pop_gmm.mean_conditional_zeta(s.xi), axis=-1)
                        for s in self.participants.values()],
                       axis=0, keepdims=False)
        # = sum of mean individual mixture weights, given xi
        logger.debug(f'{repr(g_name)}: Before pruning: w_sum = '
                     + np.array2string(w_sum, precision=2, suppress_small=True))
        if np.any(np.logical_and(min_weight < w_sum, w_sum <= 2.5)):
            logger.warning(f'Group {repr(g_name)}: *** Some GMM component(s) has TWO or LESS participants. '
                           'Variance estimates may be unreliable. ***')
        keep = min_weight < w_sum
        n_keep = np.sum(keep)
        n_total = len(keep)
        self.pop_gmm.prune(keep)
        # --- make logger message:
        msg = f'Group {repr(g_name)}:\t'
        if n_keep < n_total:
            msg += f'Model pruned to {n_keep} active mixture component(s) out of initially {n_total}'
        else:
            msg += f'All initial {n_total} mixture component(s) still active.'
        if len(self.pop_gmm.comp) > 1:
            msg += (f'\nParticipant GMM weights =\n\t'
                    + '\n\t'.join((str(s_name) + ':\t'
                                   + np.array_str(np.mean(self.pop_gmm.mean_conditional_zeta(s_model.xi),
                                                          axis=-1),
                                                  precision=2,
                                                  suppress_small=True)
                                   for (s_name, s_model) in self.participants.items())))
        logger.info(msg)

    # ---------------------------- make final results for display:
    def predictive_population_ind(self):
        """Predictive probability-distribution for a Random Individual
        in the population represented by self
        :return: a PredictivePopulationModel object
        """
        return self.pop_gmm.predictive_population_ind()

    def predictive_population_mean(self):
        """Predictive probability-distribution for MEAN parameter vector
        in the population represented by self
        :return: a PredictivePopulationModel object
        """
        return self.pop_gmm.predictive_population_mean()

# ------------------------------------- display functions
    def attribute_grade_count(self, a, groupby=None):
        """Collect table of ordinal grade counts for ONE attribute,
        summed across all participants, optionally subdivided by situation.
        :param a: selected attribute key
        :param groupby: (optional) single situation dimension or sequence of such keys
            for which separate attribute-counts are calculated.
            Counts are summed across any OTHER situation dimensions.
        :return: a pd.DataFrame object with all grade counts,
            with one row for each (*groupby) combination
            and one column for each grade category
        """
        return sum((s_model.attribute_grade_count(a, groupby=groupby)
                    for s_model in self.participants.values()))

    def rvs_grade_count(self, a, groupby=None):
        """Collect table of predicted grade counts for ONE attribute,
        summed across all participants, optionally subdivided by situation.
        :param a: selected attribute key
        :param groupby: (optional) single situation dimension or sequence of such keys
            for which separate attribute-counts are calculated.
            Counts are summed across any OTHER situation dimensions.
        :return: a pd.DataFrame object with all grade counts,
            with one row for each (sample, *groupby) combination
            and one column for each grade category
        """

        return sum((s_model.rvs_grade_count(a, groupby=groupby)
                    for s_model in self.participants.values()))


# -------------------------------------------------------------------
class PopulationModel:
    """A Gaussian Mixture Model (GMM) representing the parameter distribution in
    a population of participants from which participants in ONE group are recruited.
    The GMM is Bayesian, implemented by random-variable properties
    mix_weight = a MixWeight object = Dirichlet-distributed mixture weights,
    comp = a list of gauss_gamma.GaussianRV objects.
    """
    def __init__(self, base, mix_weight, comp, rng):
        """
        :param base: single common EmaParamBase object, used by all model parts
        :param mix_weight: a single MixtureWeight(DirichletVector) instance,
            with one element for each element of base.comp
        :param comp: list of GaussianRV instances,
            each representing ONE mixture component
            for parameter vector xi, in the (sub-)population represented by self
        :param rng: random Generator instance
        """
        self.base = base
        self.mix_weight = mix_weight
        self.comp = comp
        self.rng = rng

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                f'\n\tmix_weight={repr(self.mix_weight)},'
                f'\n\tcomp= {len(self.comp)} mixture components)'
                )

    @classmethod
    def initialize(cls, n_comp, participants, base, rng):
        """Crude initial group model given group EMA data
        :param n_comp: integer number of GMM components
        :param participants: list with EmaRespondentModel instances for group participants,
            only crudely initialized
        :param base: single common EmaParamBase object, used by all model parts
        :param rng: random Generator for this group
        :return: a cls instance crudely initialized
        """
        mix_weight = MixtureWeight(alpha=np.ones(n_comp) * PRIOR_MIXTURE_CONC,
                                   rng=rng)
        # -> equal mean weights for all mixture components
        comp = [copy.deepcopy(base.comp_prior)
                for _ in range(n_comp)]
        self = cls(base, mix_weight, comp, rng)
        self._init_comp(participants)
        # Mixture weights will be adapted later in the general VI procedure.
        return self

    def _init_comp(self, participants):
        """Initialize Gaussian mixture components to make them distinctly separated,
        using only initialized values for all participant.xi.
        This is a necessary starting point for VI learning.
        :param participants: dict with (participant_id, EmaRespondentModel-object) elements
            crudely initialized
        :return: None

        Method: pull self.comp elements apart by random method like k-means++
        that tends to maximize separation between mixture components.
        """
        def distance(x, c):
            """Square-distance from given samples to ONE mixture component,
            as estimated by component logpdf.
            :param x: 2D array of sample row vectors that might be drawn from c
            :param c: ONE mixture component in self.base.comp
            :return: d = 1D array with non-negative distance measures
                d[n] = distance from x[n] to c >= 0.
                len(d) == x.shape[0]
            """
            d = - c.mean_logpdf(x)
            return d - np.amin(d)

        def weight_by_dist(d):
            """Crude weight vector estimated from given distance measures
            :param d: 1D array with non-negative distances
            :return: w = 1D array with weights,
                with ONE large element randomly chosen with probability prop.to d2,
                and all other elements small and equal.

                Purpose: same as in kmeans++:
                    Tends to push current comp element FAR away
                    from all other comp elements already initialized.
                w.shape == d.shape
            """
            w = np.full_like(d, 1. / len(d))
            # ALL samples jointly contribute with weight equiv to ONE sample
            i = self.rng.choice(len(d), p=d / sum(d))
            w[i] = len(d) / len(self.comp)
            # total weight of all samples divided uniformly across components
            # placed on the single selected i-th sample point
            return w

        # --------------------------------------------------------
        xi = np.array([np.mean(s_model.xi, axis=0)
                       for s_model in participants.values()])
        xi2 = np.array([np.mean(s_model.xi**2, axis=0)
                        for s_model in participants.values()])
        xi_d = np.full(len(xi), np.finfo(float).max / len(xi) / 2)
        # = very large numbers that can still be normalized to sum == 1.
        for c_i in self.comp:
            c_i.adapt(xi, xi2, w=weight_by_dist(xi_d), prior=self.base.comp_prior)
            xi_d = np.minimum(xi_d, distance(xi, c_i))
            # xi_d[n] = MIN distance from xi[n] to ALL already initialized components

    def adapt(self, g_name, participants):
        """One VI adaptation step for all model parameters
        :param g_name: group id label for logger output
        :param participants: dict with current (participant_id, EmaRespondentModel) elements
        :return: ll = scalar VI lower bound to data log-likelihood,
            incl. negative contributions for parameter KLdiv re priors

        NOTE: All contributions to VI log-likelihood
        are calculated AFTER all updates of factorized parameter distributions
        because all must use the SAME parameter distribution
        """
        ll_weights = self._adapt_mix_weight(participants)
        # = -KLdiv{ q(self.mix_weight) || prior.mix_weight}
        ll_comp = self._adapt_comp(g_name, participants)
        # ll_comp[c] = -KLdiv(current q(mu_c, Lambda_c) || prior(mu_c, Lambda_c)), for
        # q = new adapted comp, p = prior_comp model, for c-th mixture component
        # Leijon doc eq:CalcLL
        sum_ll_comp = sum(ll_comp)
        logger.debug(f'{repr(g_name)}.pop_gmm: '
                     + f'\n\tll_weights= {ll_weights:.4f} '
                     + f'\n\tmix_weight.alpha= '
                     + np.array2string(self.mix_weight.alpha,
                                       precision=2,
                                       suppress_small=True)
                     + f'\n\tcomp: -KLdiv= {sum_ll_comp: .3f} = sum'
                     + np.array_str(np.array(ll_comp),
                                    precision=2,
                                    suppress_small=True))
        return ll_weights + sum_ll_comp

    def _adapt_comp(self, g_name, participants):
        """One VI update for all GMM components in self.comp
        :param g_name: group id label for logger output
        :param participants: dict with current (participant_id, EmaRespondentModel) elements
        :return: ll = sum_c (-KLdiv re prior) across self.base.comp[c]
        """
        (m_zeta, m_zeta_xi, m_zeta_xi2) = ([], [], [])
        # = accumulators for <zeta>, <zeta * xi>, <zeta * xi**2>
        for s in participants.values():
            zeta = self.mean_conditional_zeta(s.xi)
            # zeta[c, ...] = E{zeta_c | s.xi[...]}; zeta_c= binary component indicator
            # for ...-th sample of participant s to be generated by c-th mixture component.
            n_xi = s.xi.shape[0]  # = zeta.shape[1]
            m_zeta.append(np.mean(zeta, axis=-1))
            m_zeta_xi.append(np.dot(zeta, s.xi) / n_xi)
            m_zeta_xi2.append(np.dot(zeta, s.xi**2) / n_xi)
        # m_zeta[s][c] = < zeta_cs >_s.xi
        # m_zeta_xi[s][c, :] = < zeta_cs * s.xi >_s.xi
        # m_zeta_xi2[s][c, :] = < zeta_cs * s.xi**2 >_s.xi
        ll = [c.adapt(xi_c, xi2_c, w_c, prior=self.base.comp_prior)
              for (c, xi_c, xi2_c, w_c) in zip(self.comp,
                                               np.array(m_zeta_xi).transpose((1, 0, 2)),
                                               np.array(m_zeta_xi2).transpose((1, 0, 2)),
                                               np.array(m_zeta).T)]
        # = list of -KLdiv(self.comp[c] || self.base.comp_prior)
        return ll

    def _adapt_mix_weight(self, participants):
        """One update step of properties self.mix_weight,
        :return: ll = - KLdiv(self.mix_weight || prior.mix_weight), after updated mix_weight
        """
        self.mix_weight.alpha = (np.sum([np.mean(self.mean_conditional_zeta(s.xi),
                                                 axis=1)
                                         for s in participants.values()],
                                        axis=0)
                                 + PRIOR_MIXTURE_CONC)
        # = Leijon doc eq:PosteriorV
        prior_alpha = PRIOR_MIXTURE_CONC * np.ones(len(self.comp))
        # *** might store a complete prior GMM in self.base,
        # instead of generating the DirichletVector here ?
        return - self.mix_weight.relative_entropy(DirichletVector(alpha=prior_alpha))

    def prune(self, keep):
        """Prune model to keep only active mixture components
        :param keep: boolean array indicating mixture components to keep
        :return: None
        Result: all model components pruned consistently
        """
        self.mix_weight.alpha = self.mix_weight.alpha[keep]
        del_index = list(np.arange(len(keep), dtype=int)[np.logical_not(keep)])
        del_index.reverse()
        # Must delete in reverse index order to avoid IndexError
        for i in del_index:
            del self.comp[i]

    # ----------------- methods used as prior for EmaRespondentModel instances:
    def logpdf(self, xi):
        """Mean log pdf of any candidate participant parameter array,
        given current GMM defined by self.mix_weight and self.comp
        averaged across self.comp parameters and self.mix_weight
        :param xi: array with parameter sample vector(s)
            xi[..., j] = ...-th sample of j-th individual parameter,
        :return: lp = array with
            lp[...] = E_self{ln p(xi[..., :] | self)}
            lp.shape == xi.shape[:-1]

        Method: Leijon doc eq:LogProbXi, prior part
        """
        return logsumexp(self.log_responsibility(xi), axis=0)

    def d_logpdf(self, xi):
        """gradient of self.logpdf(xi) w.r.t. xi
        :param xi: 2D array with parameter sample vector(s)
            xi[..., j] = ...-th sample of j-th individual parameter,
        :return: dlp = array with
            dlp[..., j] = d log p(xi[..., :] | self) / d xi[..., j]
            lp.shape == xi.shape
        """
        return np.sum(softmax(self.log_responsibility(xi)[..., None],
                              axis=0) * self.d_log_responsibility(xi),
                      axis=0)

    def mean_conditional_zeta(self, xi):
        """Expected value of mixture-component indicator zeta,
         given any candidate participant parameter vector(s)
        :param xi: 1D or 2D array with
            xi[..., :] = ...-th parameter sample vector
         :return: w = array, normalized for sum(w) == 1.
             w[c, ...] = P{xi[..., :] generated by c-th mixture component}
             w.shape == (len(self.comp), *xi.shape[:-1])
        """
        return softmax(self.log_responsibility(xi), axis=0)

    def mean_marginal_zeta(self, xi):  # *** needed ?
        """Marginal expected value of mixture-component indicator zeta,
         given any candidate participant parameter vector(s),
         averaged across all xi samples
        :param xi: 2D array with
            xi[s, :] = s-th parameter sample vector
         :return: w = 1D array, normalized for sum(w) == 1.
             w[c] = mean_s(P{xi[s, :] generated by c-th mixture component})
             w.shape == (len(self.comp),)
        """
        return np.mean(self.mean_conditional_zeta(xi), axis=1)

    def log_responsibility(self, xi):
        """Expected log pdf for any candidate participant parameters,
        for each comp in current GMM represented by self.mix_weight and self.comp
        :param xi: 2D array with parameter sample vector(s)
            xi[..., j] = ...-th sample of j-th individual parameter,
        :return: lr = array with
            lr[c, ...] = E{ log P[ xi[..., :] | self.comp[c] }
            lr.shape == (len(self.comp), *xi.shape[:-1])

        Method: Leijon doc l_c(xi) from eq:LogProbXiZeta
        """
        return np.array([c.mean_logpdf(xi) + lv_c
                         for (c, lv_c) in zip(self.comp,
                                              self.mix_weight.mean_log)
                         ])

    def d_log_responsibility(self, xi):
        """Gradient of self.log_responsibility(xi) w.r.t. xi
        :param xi: 2D array with parameter sample vector(s)
            xi[..., j] = ...-th sample of j-th individual parameter,
        :return: dlr = array with
            dlr[c, ..., j] = d self.log_responsibility(xi)[c, ...] / d xi[..., j]
            dlr.shape == (len(self.comp), *xi.shape)
        """
        return np.array([c.grad_mean_logpdf(xi)
                         for c in self.comp])

    # ---------------------------- final results for display:
    def predictive_population_ind(self):
        """Predictive probability-distribution for Random Individual
        population represented by self
        :return: a PredictivePopulationModel object

        Method: report eq:PredictiveSubPopulation
        """
        comp = [c_m.predictive(rng=self.rng) for c_m in self.comp]
        return PredictivePopulationModel(self.base, comp, self.mix_weight.mean, self.rng)

    def predictive_population_mean(self):
        """Predictive probability-distribution for MEAN parameter vector
        in population represented by self
        :return: a PredictivePopulationModel object

        Method: report eq:PredictivePopulation
        """
        comp = [c_m.mean.predictive(rng=self.rng) for c_m in self.comp]
        return PredictivePopulationModel(self.base, comp, self.mix_weight.mean, self.rng)


# ------------------------------------------------------------------
class PredictivePopulationModel:
    """Help class defining a non-Bayesian predictive mixture model
    for parameter distribution in a population,
    derived from existing trained model components
    """
    def __init__(self, base, comp, w, rng):
        """
        :param base: ref to common EmaParamBase object
        :param comp: list of predictive mixture component models
            NOT same as original base.comp
        :param w: 1D array with mixture weight values
        :param rng: random Generator instance
        """
        self.base = base
        self.comp = comp
        self.w = w
        self.rng = rng

    @property
    def mean(self):
        """Mean of parameter vector, given population mixture,
        averaged across mixture components
        and across posterior distribution of component concentration params.
        :return: 1D array
        """
        return np.dot(self.w, [c_m.mean for c_m in self.comp])

    def rvs(self, size=N_SAMPLES):
        """Generate random probability-profile samples from self
        :param size: integer number of sample vectors
        :return: xi = 2D array of parameter-vector samples
            xi[s, :] = s-th sample parameter vector,
                with parameter sub-types as specified by self.base
        """
        n_comp = len(self.comp)
        comp_ind = self.rng.choice(n_comp, p=self.w, size=size)
        # = array of random comp indices
        n_comp_samples = [np.sum(comp_ind == i) for i in range(n_comp)]
        # = n samples to generate from each mixture component
        xi = np.concatenate([c.rvs(size=n_c)
                             for (n_c, c) in zip(n_comp_samples, self.comp)],
                            axis=0)
        self.rng.shuffle(xi, axis=0)
        return xi


# ------------------------------ help functions for Pool multi-tasking:
def _pool_size(n):
    """Estimate number of Pool sub-processes
    :param n: total number of independent jobs to share between processes
    :return: n_processes
    """
    # NOTE: cpu_count() returns n of LOGICAL cpu cores.
    return min(os.cpu_count(), n)


def _adapt_respondent(s_item):
    """dispatch function call to given object.adapt(...)
    for use by multi-processing Pool.imap() or non-pool map()
    :param s_item: tuple(s_id, s_model), where
        s_id is object id, and s_model has an adapt method
        Here, s_model is always an ema_respondent.EmaRespondentModel instance
    :return: tuple(s_id, s_model_adapted)
    """
    (s_id, s_model) = s_item
    return s_id, s_model.adapt(s_id)



"""This module defines classes to access and store recorded EMA data,
and methods and functions to read and write such data.

Each EMA Record includes nominal and ordinal elements, defining
* a participant ID label,
* a single SITUATION specified by selected categories in ONE or MORE Situation Dimension(s),
* ordinal RATING(s) for ZERO, ONE, or MORE perceptual ATTRIBUTE(s) in the current situation.
* (optionally) group category label(s) for this participant

*** Class Overview:

EmaFrame: defines study layout and category labels of data in each EMA Record.
    EmaFrame properties can also define selection criteria
    for a subset of data to be included for analysis.

EmaDataSet: container of all EMA data to be used as input for statistical analysis.

*** File Formats:

Data may be stored in various table-style file formats allowed by package pandas,
e.g., xlsx, csv, odt, etc.
Each data file may include EMA records from ONE or SEVERAL participants.

The participant id may be stored in a designated column of the table,
or the file name may be used as participant id,
or (in Excel-type files) the participant may be identified by the sheet name.

Group category(-ies) may be defined in table column(s),
OR defined by sub-string(s) of the absolute path for each file.

A collected EMA data set can be saved in various file formats, similar to input files.

*** Input Data Files:

All input files from an experiment must be stored in ONE directory tree.

If a group dimension is specified by the file path, the path must include a sub-string
combining the group dimension and one category,
for example, 'Age_old', if the group dimension 'Age' has allowed categories 'young' and 'old'.

Each participant in the same group MUST have a unique participant ID.

Different files may include data for the same participant,
e.g., results obtained in different test phases,
or simply for additional sets of EMA records from the same participant.

Participants in different groups may have the same participant ID values,
because the groups are separated anyway,
but normally the participant IDs should be unique across all groups.

*** Example Directory Tree:

Assume we have data files in the following directory structure:
~/ema_study / Age_old / Gender_male, containing files Test_EMA_64.xlsx and Response_Data_LAB_64.xlsx
    with data to be analyzed for group key= ('old', 'male') in group dimensions ('Age', 'Gender').
~/ema_study / Age_old / Gender_female, containing files Test_EMA_64.xlsx and Data_EMA_65.xlsx
~/ema_study / Age_young / Gender_male,  containing files Test_EMA_64.xlsx and EMA_65.xlsx
~/ema_study / Age_young / Gender_female, containing files Test_EMA_64.xlsx and EMA_65.xlsx

Four separate groups may then be defined by dimensions 'Age' and 'Gender',
and the analysis may be restricted to only use data in files with names including 'EMA_64'.

*** Accessing Input Data for Analysis:
*1: Create an EmaFrame object defining the experimental layout, e.g., as:

emf = EmaFrame.setup(situations={'CoSS': [f'C{i}' for i in range(1, 8)],
                                 'Important': ('Slightly', 'Medium', 'Very'),
                                 },  # nominal variables
                     attributes={'Speech': ('Very Hard', 'Fairly Hard', 'Fairly Easy','Very Easy')},
                     groups={'Age': ('young', 'old'),
                             'Gender': ('female', 'male'),
                             'Test': ('EMA_64',)},
        )
NOTE: String CASE is always distinctive, i.e., 'Male' and 'male' are DIFFERENT categories.

*2: Load all test results into an EmaDataSet object:

ds = EmaDataSet.load(emf, path='~/ema_study',
                    fmt='xlsx',
                    participant='sheet',  # xlsx sheet title is used as participant ID
                    path_groups=['Age', 'Gender', 'Test']  # group dimensions identified by file path
                    )
The parameter emf is an EmaFrame object that defines all variables to be analyzed,
and the other arguments specify where these variables are to be found.

The resulting data set includes exactly FOUR combinations of 'Age' and 'Gender',
with an extra dimension 'Test' that has only one allowed category.
The object ds can now be used as input for analysis.

*** Selecting Subsets of Data for Analysis:
It is possible to define a data set using only a subset of the data files in the given directory tree.
For example, assume we want to analyze only TWO groups, old males, and old females,
and only data from group dimension 'Test' with the single category 'EMA_64',
and only responses for Situation dimension 'CoSS'.

Then we must define a new EmaFrame object, and load only a subset of group data:

emf = EmaFrame.setup(situations={'CoSS': [f'C{i}' for i in range(1, 8)],
                                 },  # nominal variables
                     attributes={'Speech': ('Very Hard', 'Fairly Hard', 'Fairly Easy','Very Easy')},
                     groups={'Age': ('old',),
                             'Gender': ('female', 'male'),
                             'Test': ('EMA_64',)}
                    )
ds = EmaDataSet.load(emf, path='~/ema_study',
                    fmt='xlsx',
                    participant='sheet',
                    path_groups=['Age', 'Gender', 'Test'])

In case all input data have been collected in a single file in the data directory,
the file must include columns with header 'Age', 'Gender',
and one column for the participant ID, with header, e.g., 'Test Subject'.
Then the data may be accessed as

ds = EmaDataSet.load(emf, path='~/ema_study',
                    fmt='xlsx',
                    participant='Test Subject',  # column header
                    path_groups=['Test']  # read only file paths including 'Test_EMA_64'
                    )
This data set will include TWO groups named ('old', 'female', EMA_64), and ('old', 'male', EMA_64)


*** Version History:
* Version 1.0.0:
2023-04-25, Group categories specified in EmaFrame.setup(...)
            EmaFrame.filter(...) does all data checking.
            EmaDataSet.load(...) allows group category in file path.
            EmaDataSet.save(...) can save separate group/participant files, or one big file.
            EmaDataSet.groups: keys include only group categories, NOT group "dimensions".

* Version 0.9.6:
2023-04-11, bugfix EmaDataSet.join_df, .attribute_count, .attribute_mean

* Version 0.9.4:
2022-11-06: Fix to avoid FutureWarning in EmaDataSet.attribute_grade_mean

* Version 0.9.3:
2022-08-22, EmaDataSet.load(), .save() safer for pandas read, in case empty phase_key
2022-08-16, changed EmaFrame.situations -> situation_dtypes
2022-08-16, changed EmaFrame.attribute_grades -> attribute_dtypes
2022-08-16, new EmaFrame.setup() method, for clarity, separate from __init__
2022-07-27, changed 'subject' -> 'participant'
2022-07-13, renamed EmaFrame.scenarios -> situations, and other methods consistently
2022-07-13, changed EmaDataSet method names: attribute_grade_count, attribute_grade_mean
2022-07-11, minor bugfix in EmaDataSet.attribute_grade_distribution

* Version 0.9.2:
2022-06-16, minor fix in EmaDataSet.mean_attribute_table, nap_table
2022-06-03, changed variable name stage -> phase everywhere
2022-05-21, clearer logger info for valid- and missing-data input

* Version 0.9:
2022-03-17, use Pandas CategoricalDtype instances in EmaFrame situation_dtypes and attributes
2022-03-18, use Pandas DataFrame format in EmaDataSet, to allow many input file formats

* Version 0.8.3:
2022-03-08, minor fix for FileReadError error message

* Version 0.8.1:
2021-02-27, fix EmaDataSet.load(), _gen_group_file_paths(), _groups(), for case NO groupby

* Version 0.5.1:
2021-11-26, EmaDataSet.load warning for input argument problems

* Version 0.5:
2021-10-15, first functional version
2021-11-18, groupby moved from EmaFrame -> EmaDataSet.load
2021-11-20, EmaDataSet.ensure_complete
2021-11-23, Group dir name MUST include both (g_factor, g_cat), e.g., 'Age_old'
2021-11-xx, allow empty attribute_grades
"""
# *** Future:
# *** EmaDataSet.initialize + add method ? load = initialize + add

import numpy as np
from pathlib import Path
import pandas as pd

from itertools import product, chain
import logging

from EmaCalc.ema_file import ema_gen, Table, FileReadError, FileWriteError
from EmaCalc.ema_nap import nap_pandas

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# ------------------------------------------------------------------
class EmaFrame:
    """Defines variable names and categories of all data elements
    to be analyzed in an EMA study.
    The allowed categories of each EMA variable is stored as a pd.CategoricalDtype instance.
    """
    def __init__(self, situation_dtypes, phase_key, ordinal_scales, attribute_scales, group_dtypes):
        """
        :param situation_dtypes: dict with elements (dimension, dtype), where
            dimension is a string label identifying one situation "dimension",
            dtype is a pd.CategoricalDtype defining NOMINAL categories within this dimension.
        :param phase_key: situation key (dimension name) for the test phase,
            which MUST be the FIRST element of situation_dtypes.
        :param ordinal_scales: dict with elements (scale_id: dtype), where
            dtype is a pd.CategoricalDtype defining ORDINAL categories for the scale.
        :param attribute_scales: dict with elements (attribute_key, scale_id), defining
            the scale used by each attribute.
            Note: Separate attributes may use the SAME ordinal scale,
                if so specified by the researcher.
        :param group_dtypes: dict with elements (dimension, dtype), where
            dimension is a string label identifying one grouping "dimension", e.g., 'Age'
            dtype is a pd.CategoricalDtype defining NOMINAL categories within this dimension.
        """
        self.situation_dtypes = situation_dtypes
        self.phase_key = phase_key
        self.ordinal_scales = ordinal_scales
        self.attribute_scales = attribute_scales
        self.group_dtypes = group_dtypes

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t\t' +
                ',\n\t\t'.join(f'{key}={repr(v)}'
                               for (key, v) in vars(self).items()) +
                '\n\t\t)')

    @classmethod
    def setup(cls, situations=None, phase_key='Phase', attributes=None, groups=None):
        """Create the EmaFrame object defining all EMA variables to be analyzed.
        :param situations: (optional) dict or iterable with elements (dimension, category_list), where
            dimension is a string label identifying one situation "dimension",
            category_list is an iterable of labels for NOMINAL categories within this dimension.
        :param attributes: (optional) dict or iterable with elements (attribute, grades),
            attribute is string id of a rated perceptual attribute,
            grades is an iterable with ORDINAL categories, strings or integer.
        :param phase_key: (optional) situation key for the test phase, with
            situation_dtypes[phase_key] = list of test phases (e.g., before vs after treatment),
                specified by experimenter, i.e., NOT given as an EMA response
            situation_dtypes[phase_key] is automatically added with a SINGLE value,
            if not already defined in given situation_dtypes.
        :param groups: (optional) dict or iterable with elements (dimension, category_list), where
            dimension is a string label identifying one grouping "dimension", e.g., 'Age' or 'Gender',
            category_list is an iterable of labels for NOMINAL categories within this dimension.

        NOTE: situation_dtypes and attribute_grades and groups may define a subset of
            data in input data files, if not all variants are to be analyzed.
            Categories MUST be specified EXACTLY as stored in data files,
            case-sensitive, i.e., 'A', and 'a' are different categories.

            If needed, the EmaDataSet.load(...) method allows
            argument 'converters' to pandas reader function,
            defining a dict with function(s) to make saved data fields agree with pre-defined categories.
            Column headers in the data files may also be re-named, if needed,
            as specified by argument rename_cols to the EmaDataSet.load(...) method.
        """
        if situations is None:
            situations = dict()
        else:
            situations = dict(situations)
        if phase_key not in situations.keys():
            situations[phase_key] = ('',)  # just a single category
        situations = dict((sit, pd.CategoricalDtype(categories=sit_cats,
                                                    ordered=False))
                          for (sit, sit_cats) in situations.items())
        # re-order to ensure first situation key == self.phase_key:
        phase_dict = {phase_key: situations.pop(phase_key)}
        situations = phase_dict | situations  # **** requires python >= 3.9

        if attributes is None:
            attributes = dict()
        else:
            attributes = dict(attributes)
        # NOTE: some attributes.values() may be IDENTICAL objects, not only equal.
        ordinal_scales = {id(a_cats): pd.CategoricalDtype(categories=a_cats,
                                                          ordered=True)
                          for (a, a_cats) in attributes.items()}
        # including only UNIQUE scales, each possibly tied to more than one attribute
        attribute_scales = dict((a, id(a_cats))
                                for (a, a_cats) in attributes.items())
        if groups is None:
            groups = dict()  # {'': ['']}  # Empty?, ONE un-named group with all participants
        groups = dict((g, pd.CategoricalDtype(categories=g_cats,
                                             ordered=False))
                     for (g, g_cats) in groups.items())
        return cls(situations, phase_key, ordinal_scales, attribute_scales, group_dtypes=groups)

    @property
    def tied_response_scales(self):
        """Some ordinal scale tied to more than one attribute scale"""
        return len(self.ordinal_scales) < len(self.attribute_scales)

    @property
    def attribute_dtypes(self):
        """Mapping attribute key -> dtype object
        :return: dict with elements (a_key, a_dtype),
            where a_dtype is a pd.CategoricalDtype object
        """
        return {a_key: self.ordinal_scales[scale_key]
                for (a_key, scale_key) in self.attribute_scales.items()}

    @property
    def scale_attributes(self):
        """Mapping scale key -> list of attributes sharing the same scale
        i.e., inverse to mapping self.attribute_scales
        :return: dict with elements (s_id, a_keys), such that
            self.ordinal_scales[s_id] is shared by all attributes in a_keys
        """
        return {s_id: [a_key for (a_key, a_scale) in self.attribute_scales.items()
                       if s_id == a_scale]
                for s_id in self.ordinal_scales.keys()}

    @property
    def dtypes(self):
        """
        :return: dict with all defined ema variables and their dtypes
        """
        return self.situation_dtypes | self.attribute_dtypes | self.group_dtypes

    @property
    def situation_shape(self):
        """tuple with number of nominal categories for each situation dimension"""
        return tuple(len(sit_dtype.categories)
                     for sit_dtype in self.situation_dtypes.values())

    @property
    def rating_shape(self):
        """tuple with number of ordinal response levels for each attribute
        """
        return tuple(len(r_cat.categories)
                     for r_cat in self.attribute_dtypes.values())

    @property
    def n_phases(self):
        # == situation_shape[0]
        return len(self.situation_dtypes[self.phase_key].categories)

    def group_head(self):
        return tuple(self.group_dtypes.keys())

    def required_vars(self):
        return list(chain(self.situation_dtypes.keys(),
                          self.attribute_dtypes.keys(),
                          self.group_dtypes.keys()))

    def filter(self, ema, participant):
        """Check and filter EMA data for ONE participant, to ensure that
        it includes the required columns, with required data types.
        :param ema: a pd.DataFrame instance
        :param participant: a required column name
        :return: a pd.DataFrame instance with complete data
            or an empty DataFrame if no usable EMA records were found
        """
        # *** check and correct for Phase column here ? ***********
        required_col = self.required_vars() + [participant]
        required_dtypes = self.dtypes | {participant: object}  # *** string?
        try:
            ema = ema[required_col]  # *** do not drop columns, only TEST ? ****
            ema = ema.astype(required_dtypes, errors='raise')
            # *** this accepts NaN, but sets NaN if cell contents not in defined categories ***
            # *** delete rows with NaN not needed? NaNs excluded anyway by DataFrame.value_counts()
        except KeyError as e:
            raise FileReadError(f'Some missing required data column(s). Error {e}')
        except ValueError as e:
            raise FileReadError(f'Incompatible data type. Error {e}')
        return ema

    def count_situations(self, ema):
        """Count EMA situation occurrences for analysis
        :param ema: np.DataFrame instance with all EMA records for ONE respondent,
            with columns including self.situation_dtypes.keys()
        :return: z = mD array with situation_counts
            z[k0, k1,...] = number of recordings in (k0, k1,...)-th situation category
            z.shape == self.situation_shape
        """
        # 2022-05-24, Arne Leijon: verified manually with input data
        z = ema.value_counts(subset=list(self.situation_dtypes.keys()), sort=False)
        # = pd.Series including only non-zero counts, indexed by situation or tuple(situation_dtypes)
        ind = pd.MultiIndex.from_product([sit_dtype.categories
                                          for sit_dtype in self.situation_dtypes.values()])
        z = z.reindex(index=ind, fill_value=0)
        # must reindex to include zero counts
        return np.array(z).reshape(self.situation_shape)

    def count_grades(self, a, ema):
        """Count grade occurrences for given attribute
        :param a: attribute key
        :param ema: pd.DataFrame instance with all EMA records for ONE respondent,
            with columns including all self.situation_dtypes.keys() and self.attribute_grade.keys()
        :return: y = 2D array with
            y[l, k] = number of responses at l-th ordinal level,
            given k-th <=> (k0, k1, ...)-th situation category
        """
        # 2022-05-24, verified manually with input data
        z = ema.value_counts(subset=[a] + list(self.situation_dtypes.keys()), sort=False)
        ind = pd.MultiIndex.from_product([self.attribute_dtypes[a].categories]
                                         + [sit_dtype.categories
                                            for sit_dtype in self.situation_dtypes.values()])
        z = z.reindex(index=ind, fill_value=0)
        return np.array(z).reshape((ind.levshape[0], -1))


# ------------------------------------------------------------
class EmaDataSet:
    """Container of all data input for one complete EMA study.
    """
    def __init__(self, emf, groups):
        """
        :param emf: an EmaFrame instance
        :param groups: dict with elements (group_id: group_dict), where
            group_id = tuple (g_cat_0, g_cat_1,...), identifying a population
                by one category selected for each group "dimension".
            group_dict = dict with elements (participant_id, ema_df), where
            ema_df = a pd.DataFrame instance with one column for each EMA variable,
            and one row for each EMA record.
            ema_df.shape == (n_records, n SITUATION dimensions + n ATTRIBUTES)
        """
        self.emf = emf
        self.groups = groups

    def __repr__(self):
        def sum_n_records(g_participants):
            """Total number of EMA records across all participants in group"""
            return sum(len(s_ema) for s_ema in g_participants.values())
        # ---------------------------------------------------------------
        return (self.__class__.__name__ + '(\n\t'
                + f'emf= {self.emf},\n\t'
                + 'groups= {' + '\n\t\t'
                + '\n\t\t'.join((f'{g}: {len(g_participants)} participants '
                                 + f'with {sum_n_records(g_participants)} EMA records in total,')
                                for (g, g_participants) in self.groups.items())
                + '\n\t\t})')

    # *** separate classmethod initialize, method add; load = initialize + add ?
    # *** to allow collecting data from several sources, with different file formats...

    @classmethod
    def load(cls, emf, path, fmt=None,
             participant='file',
             path_groups=None,
             group_join_str='_',
             ema_vars=None, grouping=None,
             **kwargs):
        """Create one class instance with selected data from input file(s).
        :param emf: EmaFrame instance
        :param path: string or Path defining top of directory tree containing all data files.
            Files in all subdirectories are searched, hierarchically.
        :param fmt: (optional) string with file suffix for data files.
            If None, all files are tried, so files with mixed formats can be used as input.
        :param participant: string defining where to find participant ID in a file,
            = column name, 'file', or 'sheet' if excel-type file that has 'sheet's
        :param path_groups: (optional) list with group-dimension labels
            for group categories to be specified by substrings in file paths,
            used only IF NOT specified in a column in the data file.
        :param group_join_str: (optional) string between group dimension and category in file path,
            e.g., '_' in '... / Age_old / ...'
        :param ema_vars: *** only for version warning, no longer used
        :param grouping: *** only for version warning, no longer used
        :param kwargs: (optional) any additional arguments for pandas file_reader
            e.g., rename_cols={...},  converters={...}.
        :return: a single cls object

        NOTE: Situation categories and Attribute grades and group categories must agree EXACTLY
            with the categories previously defined in emf.
            If needed, use keyword argument
            converters={column_head: convert_fcn, ...} to pandas reader function
            with converter function(s) to make saved data fields agree with pre-defined categories.
            Use keyword argument
            rename_cols={old_head: new_head, ...} to change file column headers to desired names.
        """
        def clean_up(s_ema):
            """Remove unused group and participant columns, re-index EMA records
            :param s_ema: a pd.DataFrame for ONE participant, incl group and participant id columns
                with a single RangeIndex for EMA record number
            :return: copy of s_ema, with redundant info removed
            """
            index_cols = list(emf.group_head()) + [participant]
            drop_cols = [c for c in index_cols if c in s_ema.columns]  # avoid KeyError in drop()
            s_ema = s_ema.drop(columns=drop_cols)
            s_ema.index = pd.RangeIndex(range(s_ema.shape[0]))
            # because the original index was artificially created by concat
            return s_ema

        # ---------------------------------------------------------------
        if ema_vars is not None:  # backwards incompatibility
            logger.warning('EmaCalc v. > 0.9.0: ema_vars not used to select EMA variables. '
                           + 'Using file table header instead. \n'
                           + 'Change column names by "rename_cols" argument, if needed.')
        if grouping is not None:  # backwards incompatibility
            logger.warning('EmaCalc v. > 0.9.6: group categories are defined in EmaFrame.setup().'
                           +'\nDefine only "path_groups" here, for group dim.s specified by path string.')
        if fmt is not None and fmt[0] != '.':
            fmt = '.' + fmt
        path = Path(path)
        if path_groups is None:
            path_groups = []
        # **** up to here -> classmethod initialize
        # **** following: -> add method, to allow collecting data from different file formats ?

        # v 1.0.0: collect a list of ema chunks, then concat and separate by group and participant:
        all_ema = []  # space for all input EMA tables
        path_groups = {g_key: emf.group_dtypes[g_key].categories
                       for g_key in emf.group_dtypes if g_key in path_groups}
        # = dict with (g_key, g_cats) for all g_keys that MAY be derived from file path.
        for (path_group_id, g_path) in _gen_group_file_paths(path, fmt, path_groups, group_join_str):
            # path_group_id = dict with element (group_key, group_cat) specified as g_path sub-strings
            logger.info(f'Path group {path_group_id}: Reading {g_path}')
            try:
                ema_file = ema_gen(g_path,
                                   participant=participant,
                                   **kwargs)
                for ema in ema_file:
                    # ema is a pd.DataFrame with a COPY of (possibly converted) data from file,
                    # with one column name == participant.
                    for (g_key, g_cat) in path_group_id.items():
                        if g_key not in ema.columns:
                            ema[g_key] = g_cat  # set group categories derived from path sub-string
                    if emf.n_phases == 1:  # phase-code might be unspecified in file
                        # if file rows have phase as empty string, Pandas reads it as NaN!
                        if emf.phase_key not in ema.columns:
                            ema[emf.phase_key] = emf.situation_dtypes[emf.phase_key].categories[0]
                    ema = emf.filter(ema, participant)
                    # = pd.DataFrame with all required columns, and nothing else!
                    logger.info(f'EMA chunk with {ema.shape[0]} records. '
                                + ('Some missing data. Valid data count =\n'
                                   + _table_valid(ema) if np.any(ema.isna()) else ''))
                    if not ema.empty:
                        all_ema.append(ema)
            except FileReadError as e:
                logger.warning(e)  # and just try next file
        if len(all_ema) == 0:
            raise FileReadError('No EMA data found')
        else:
            all_ema = pd.concat(all_ema, axis=0, ignore_index=True)
        # store all data in structured dict instead: **** or just leave it as a single chunk ? ***
        if len(emf.group_dtypes) == 0:
            groups = {(): {s: clean_up(s_ema)
                           for (s, s_ema) in all_ema.groupby(participant)}
                      }
        else:
            groups = {g: {s: clean_up(s_ema)
                          for (s, s_ema) in g_ema.groupby(participant)}
                      for (g, g_ema) in all_ema.groupby(list(emf.group_dtypes.keys()))}
        return cls(emf, groups)

    # def add method, to include data from new files with different layout ???

    def save(self, path, fmt='.csv', allow_over_write=False,
             participant='file',
             join_groups=False,
             group_join_str='_',
             **kwargs):
        """Save all EMA data from self.groups, either
        as a single file with EMA records for all participants in all groups,
        OR in a directory tree for groups, with one file for each group,
        OR with one separate file for each participant.
        :param path: Path or string defining the top directory where files are saved
        :param allow_over_write: (optional) boolean switch
        :param fmt: (optional) file-name extension string specifying file format
        :param join_groups: (optional) boolean switch: True -> ONE file for all groups
        :param participant: (optional) string defining where to store participant ID,
            = column name, -> all participants in a group saved in ONE file
            or 'file' -> one file for each participant, with participant ID as file name
        :param group_join_str: (optional) string between group dimension and category in file path
        :param kwargs: (optional) arguments to Table.save() or selected pandas.to_xxx()
        :return: None
        """
        def drop_empty_phase(df):
            # if Phase column contains only empty strings, drop it, do NOT save it,
            # because Pandas might read it as NaN, not as empty string
            phase_key = self.emf.phase_key
            if self.emf.n_phases == 1 and all(df[phase_key] == ''):
                df = df.drop(columns=[phase_key], inplace=False)
            return df

        # --------------------------------------------------------------
        if fmt[0] != '.':
            fmt = '.' + fmt
        path = Path(path)
        if participant == 'file':
            if join_groups:
                join_groups = False
                logger.warning('Must have join_groups = False to save participants in separate files')
        try:
            if join_groups:  # AND participants
                groups = self.join_df(participant=participant)
                groups = drop_empty_phase(groups)
                f_name = group_join_str.join(emf.group_head())
                if len(f_name) == 0:
                    f_name = 'participants'
                else:
                    f_name += '_participants'
                f_path = (path / f_name).with_suffix(fmt)
                f_path.parent.mkdir(parents=True, exist_ok=True)
                Table(groups).save(f_path, allow_over_write, **kwargs)
            else:
                for (g, group_data) in self.groups.items():
                    g_dir = group_dir_str(self.emf.group_head(), g, sep=group_join_str)
                    # = group sub-path string, empty if single un-named group
                    g_path = path / g_dir
                    if participant == 'file':
                        g_path.mkdir(parents=True, exist_ok=True)
                        for (s_id, s_df) in group_data.items():
                            s_df = drop_empty_phase(s_df)
                            p = (g_path / str(s_id)).with_suffix(fmt)  # one file per participant
                            Table(s_df).save(p, allow_over_write, **kwargs)
                    else:
                        g_table = pd.concat(group_data, axis=0,
                                            ignore_index=False,
                                            sort=False,
                                            names=[participant]
                                            )
                        g_table = drop_empty_phase(g_table)
                        f_path = g_path.with_suffix(fmt)  # last dir level becomes the file
                        g_path.parent.mkdir(parents=True, exist_ok=True)
                        Table(g_table).save(f_path, allow_over_write, **kwargs)
        except FileWriteError as e:
            raise RuntimeError(f'Could not save {self.__class__.__name__} in {repr(fmt)} format. '
                               + f'Error: {e}')

    def ensure_complete(self):
        """Check that we have at least one participant in every sub-population category,
        with at least one ema record for each participant (already checked in load method).
        :return: None

        Result:
        self.groups may be reduced:
        participants with no records are deleted,
        groups with no participants are deleted
        logger warnings for missing data.
        """
        for (g, g_participants) in self.groups.items():
            incomplete_participants = set(s for (s, s_ema) in g_participants.items()
                                      if s_ema.empty)
            for s in incomplete_participants:
                logger.warning(f'No EMA data for participant {repr(s)} in group {repr(g)}. Deleted!')
                del g_participants[s]
        incomplete_groups = set(g for (g, g_participants) in self.groups.items()
                                if len(g_participants) == 0)
        for g in incomplete_groups:
            logger.warning(f'No participants in group {repr(g)}. Deleted!')
            del self.groups[g]
        if len(self.groups) == 0:
            raise RuntimeError('No EMA data in any group.')
        for attr in self.emf.attribute_dtypes.keys():
            a_count = self.attribute_grade_count(attr)
            # = pd.DataFrame with all groups, all participants
            _check_ratings(attr, a_count)

    def join_df(self, participant='Participant'):  # **** default 'Participant' where ?
        """Join all EMA data into ONE single pd.DataFrame instance
        for all groups and all participants
        :param participant: (optional) name of column with participant ID
        :return: a single pd.DataFrame instance  **** NOT Table ?
        """
        g_dict = {g_key: pd.concat({s: s_data
                                    for (s, s_data) in g_data.items()},
                                   axis=0,
                                   sort=False,
                                   names=[participant])
                  for (g_key, g_data) in self.groups.items()}
        # if len(g_dict) == 1 and len(self.emf.group_head()) == 0:  # only ONE UN-NAMED group
        #     (g_cat, df) = g_dict.popitem()
        # else:
        #     df = pd.concat(g_dict, axis=0, names=self.emf.group_head(), sort=False)
        df = self._join_group_dict(g_dict)
        return Table(df)

    def _join_group_dict(self, g_dict):
        """Help method to concat pd.DataFrame or pd.Series instances across groups,
        with special treatment in case of single un-named group.
        Needed only internally.
        :param g_dict: dict with elements (g_key, g_data), where
            g_data is a DataFrame or pd.Series object
        :return: a single pd.DataFrame or pd.Series instance
        """
        if len(g_dict) == 1 and len(self.emf.group_head()) == 0:  # only ONE UN-NAMED group
            (g_cat, df) = g_dict.popitem()
        else:
            df = pd.concat(g_dict, axis=0, names=self.emf.group_head(), sort=False)
        return df

    def attribute_grade_count(self, a, groupby=None):
        """Collect table of ordinal grades for ONE attribute,
        for each (group, participant), optionally sub-divided by situation
        :param a: ONE selected attribute key
        :param groupby: (optional) single situation dimension or list of such dimensions
            for which separate attribute-counts are calculated.
            Counts are summed across any OTHER situation dimensions.
        :return: a pd.DataFrame object with all grade counts,
            with one row for each (group, participant, *groupby) case
            and one column for each grade category
        2023-04-11, bugfix
        """
        def s_count(s_data, a, groupby):
            """Calculate participant value_count
            :param s_data: a participant DataFrame
            :param a: attribute key
            :param groupby: list of situation dimension, possibly empty
            :return: DataFrame instance with desired value counts for attribute a
            """
            if len(groupby) == 0:
                return s_data[a].value_counts(sort=False)
            else:
                return s_data.groupby(groupby)[a].value_counts(sort=False)
        # ------------------------------------------------------------

        if groupby is None:
            groupby = []
        elif isinstance(groupby, str):
            groupby = [groupby]
        groupby = [gb for gb in groupby if gb in self.emf.situation_dtypes.keys()]
        g_dict = {g_key: pd.concat({s: s_count(s_data, a, groupby)
                                    for (s, s_data) in g_data.items()},
                                   axis=0,
                                   sort=False,
                                   names=['Participant'])  # *** -> cls property ?
                  for (g_key, g_data) in self.groups.items()}
        # if len(g_dict) == 1 and len(self.emf.group_head()) == 0:  # only ONE UN-NAMED group
        #     (g_cat, df) = g_dict.popitem()
        # else:
        #     df = pd.concat(g_dict, axis=0, names=self.emf.group_head(), sort=False)
        df = self._join_group_dict(g_dict)
        return Table(df.unstack(a))

    def attribute_grade_mean(self, a=None, groupby=None):
        """Average raw attribute grades, encoded numerically as (1,.., n_grades)
        :param a: (optional) attribute label or sequence of attribute,
            if None, include all attributes
        :param groupby: (optional) single situation dimension or iterable of such keys
            for which separate attribute-means are calculated.
            Results are aggregated across any OTHER situation dimensions.
        :return: a pd.DataFrame instance with all mean Attribute grades,
            with rows Multi-indexed for Group(s), Participant, and selected Situation dimensions.
            with one column for selected attribute(s).
        """
        def recode_attr(df, a):
            """Recode ordinal attribute grades linearly to numerical (1,...,n_grades)
            :param df: a pd.DataFrame instance
            :param a: list of attribute column names in df
            :return: None; df recoded in place
            """
            # *** allow external user-defined recoding function ?
            for a_i in a:
                c = df[a_i].array.codes.copy().astype(float)
                c[c < 0] = np.nan
                df[a_i] = c + 1

        def s_mean(s_data, a, groupby):
            """Calculate participant value_count
            :param s_data: a participant DataFrame
            :param a: attribute key or list of such keys
            :param groupby: list of situation dimension, possibly empty
            :return: DataFrame instance with desired value counts for attribute a
            """
            s_data = s_data.copy()  # avoid modifying original
            recode_attr(s_data, a)
            if len(groupby) == 0:
                return s_data[a].mean(numeric_only=True)
            else:
                return s_data.groupby(groupby)[a].mean(numeric_only=True)
        # ------------------------------------------------------------
        if a is None:
            a = list(self.emf.attribute_dtypes.keys())
        elif isinstance(a, str):
            a = [a]
        a = [a_i for a_i in a if a_i in self.emf.attribute_dtypes.keys()]
        if groupby is None:
            groupby = []
        elif isinstance(groupby, str):
            groupby = [groupby]
        groupby = [gb for gb in groupby if gb in self.emf.situation_dtypes.keys()]
        g_dict = {g_key: pd.concat({s: s_mean(s_data, a, groupby)
                                    for (s, s_data) in g_data.items()},
                                   axis=0,
                                   sort=False,
                                   names=['Participant'])  # *** -> cls property ?
                  for (g_key, g_data) in self.groups.items()}
        # if len(g_dict) == 1 and len(self.emf.group_head()) == 0:  # only ONE UN-NAMED group
        #     (g_cat, df) = g_dict.popitem()
        # else:
        #     df = pd.concat(g_dict, axis=0, names=self.emf.group_head(), sort=False)
        df = self._join_group_dict(g_dict)
        return Table(df)

    def nap_table(self, sit, nap_cat=None, a=None, groupby=None, p=0.95):
        """Calculate proportion of Non-overlapping Pairs = NAP result
        in ONE situation dimension with EXACTLY TWO categories, X and Y,
        = estimate of P(attribute grade in X < attribute grade in Y),
        given observed ordinal i.i.d. grade samples for attribute in situation_dtypes X and Y.
        :param sit: ONE situation dimension with TWO categories to be compared
        :param nap_cat: (optional) sequence of TWO categories (X, Y) in situation dimension sc.
            If None, sit MUST be categorical with exactly TWO categories.
        :param a: (optional) attribute name or list of attribute names
        :param groupby: (optional) single situation key or iterable of such keys
            for which separate NAP results are calculated.
            Results are aggregated across any OTHER situation dimensions.
        :param p: (optional) scalar confidence level for NAP result
        :return: a pd.DataFrame instance with all NAP results,
            with rows Multi-indexed for Group(s), Participant, grouping Situation-dimension(s),
            columns Multi-indexed with three NAP results for each Attribute:
            (lower conf-interval limit, point estimate, upper conf-interval limit)
        """
        if a is None:
            a = list(self.emf.attribute_dtypes.keys())
        elif isinstance(a, str):  # single attribute
            a = [a]
        a = [a_i for a_i in a
             if a_i in self.emf.attribute_dtypes.keys()]
        if groupby is None:
            groupby = []
        elif isinstance(groupby, str):
            groupby = [groupby]
        groupby = [gb for gb in groupby if gb in self.emf.situation_dtypes.keys()]
        df = self.join_df()
        g_cols = list(self.emf.group_dtypes.keys())
        if len(g_cols) == 1 and len(g_cols[0]) == 0:
            g_cols = []
        groupby = g_cols + ['Participant'] + groupby
        return Table(nap_pandas(df, col=sit, nap_cat=nap_cat,
                                group_cols=groupby, grade_cols=a, p=p))


# -------------------------------------------- module help functions

def group_dir_str(g_head, g, sep='_'):
    """Convert group id to a directory path string
    :param g_head: tuple of strings with group "dimensions"
    :param g: tuple of labels with corresponding group categories
    :param sep: string joining group dimension and category
    :return: string to be used as directory path
    """
    s = '/'.join((str(h_i) + sep + str(g_i)
                     for (h_i, g_i) in zip(g_head, g)))
    if s == sep:  # only a single group with empty head and category
        return ''
    else:
        return s


def _gen_group_file_paths(path, fmt, group_factors, sep='_'):
    """Generator of group categories and corresponding file Paths, recursively, for all groups
    :param path: Path instance defining top directory to be searched
    :param fmt: file suffix of desired files, INCL leading '.'
    :param group_factors: dict with tuples (g_factor, list of categories)
    :param sep: (optional) separator string between group dimension and category in file path
    :return: generator of tuples (group_key, file_path), where
        group_key is a dict with elements (g_key, g_cat), for all g_key in group_factors,
        file_path is a Path object to a file that may hold EMA data for the group,
        Only paths with matching group sub-strings AND desired file format are included.
    """
    if fmt is None:
        path_gen = path.glob('**/*.*')  # try all files in all subdirectories
    else:
        path_gen = path.glob('**/*' + fmt)  # try only requested file format
    for p in path_gen:
        group_key = _match_group_factors(p, group_factors, sep)
        # = empty dict if group_factors is empty dict -> OK file
        # = None if group_factors is non-empty, but NOT ALL matched in p
        if group_key is not None:
            yield (group_key, p)


def _match_group_factors(p, group_factors, sep='_'):
    """
    :param p: path to be checked
    :param group_factors: dict with elements (g_key, g_cats),
        where g_cats is a list of allowed categories fpr group dimension g_key
    :param sep: (optional) separator string between group dimension and category in file path
    :return: g_cats = dict with elements (g_key, g_cat), for all g_key in group_factors,
        IFF all matching patterns, e.g., 'Age_old' for g_key='Age', g_cat = 'old'
        were found in p.
        Otherwise, None, if not all group_factors were matched.
    """
    def find_in_string(s, k, cats):
        for c in cats:
            if 0 <= s.find(str(k) + sep + str(c)):
                return c
        return None

    # -------------------------------------------------
    g_keys = dict()
    for (g_key, g_cats) in group_factors.items():
        g_cat = find_in_string(str(p), g_key, g_cats)
        if g_cat is None:
            return None
        else:
            g_keys[g_key] = g_cat
    return g_keys


def _table_valid(ema: pd.DataFrame):
    """Count valid data elements for all columns
    :param ema: Pandas.DataFrame instance with input EMA data
    :return: table string for logger output
    """
    return pd.DataFrame([ema.count()]).to_string(index=False)


def _check_ratings(a, a_count):
    """Warning about zero rating counts in some categories
    :param a: attribute key
    :param a_count: pd.DataFrame with count distribution for this attribute
        one row for each (group, participant), summed across Situations
    :return: None
    """
    max_zero = 0.5  # proportion of all participants
    n_rows = a_count.shape[0]
    zero_participants = np.sum(a_count.to_numpy() == 0, axis=0)
    if np.any(zero_participants == n_rows):
        logger.warning(f'Attribute {a}: Some grades unused by ALL participants! '
                       + 'Consider merging grades?\n\t'
                       + f'{a} grades=\n'
                       + a_count.to_string())
    elif np.any(zero_participants > max_zero * n_rows):
        logger.warning(f'Attribute {a}: Some grades unused by some participants! '
                       + 'Consider merging grades?')


# -------------------------------------------- TEST:
if __name__ == '__main__':
    import ema_logging
    # ------------------------ Set up working directory and result logging:
    work_path = Path.home() / 'Documents' / 'EMA_sim'  # or whatever...
    data_path = work_path / 'data'  # to use simulation data generated by run_sim.py
    # result_path = work_path / 'result'  # or whatever

    # model_file = 'test_ema_model.pkl'  # name of saved model file (if saved)

    ema_logging.setup()

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)

    # ------ 1: Define Experimental Framework: Situations, Attributes, and Grades

    # NOTE: This example uses data generated by template script run_sim.py
    # Edit as needed for any other EMA data source

    sim_situations = {# 'phase': ('',),  # only ONE Test phase with empty label
                     'HA': ('A', 'B'),  # Two Hearing-aid programs
                     'CoSS': [f'C{i}' for i in range(1, 8)],    # Seven CoSS categories
                     }  # nominal variables, same for all (Sub-)Populations
    # NOTE: First situation dimension is always phase, even if only ONE category
    # User may set arbitrary phase_key label
    # Dimension 'phase' may be omitted, if only one category

    emf = EmaFrame.setup(situations=sim_situations,
                         phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         groups={'Age': ('young', 'old'),
                                 # 'Gender': ('M', 'F'),
                                 }
                         )

    print('emf=\n', emf)
    print(f'emf.n_phases= {emf.n_phases}')
    print(f'emf.situation_shape= {emf.situation_shape}')
    print(f'emf.rating_shape= {emf.rating_shape}')

    ds = EmaDataSet.load(emf, data_path, fmt='csv',
                         participant='file',
                         path_groups=['Age', 'Gender'],
                         dtype={'CoSS': 'string'})
    print('ds= ', ds)

    test = ds.attribute_grade_count(a='Speech', groupby=('HA', 'CoSS'))
    test.to_string(work_path / 'test_attribute_count.txt')
    print('rating_count=\n', test)

    test = ds.attribute_grade_mean(groupby=('HA', 'CoSS'))
    test.to_string(work_path / 'test_attribute_mean.txt')
    print('mean_rating=\n', test)

    nap = ds.nap_table(sit='HA', groupby=('CoSS',))
    nap.to_string(work_path / 'test_nap_table.txt', float_format='%.3f')
    print('NAP(HA B > A)=\n', nap)

    # -------------------- test EmaDataSet.save
    test_path = work_path / 'test_save'
    ds.save(test_path, fmt='csv', participant='file', join_groups=False, group_join_str='_')
    print(f'Data set saved in {test_path}')

    # -------------------------------------TEST zero group dimensions
    emf = EmaFrame.setup(situations=sim_situations,
                         phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         # groups={'Age': ('young', 'old'),
                         #         # 'Gender': ('M', 'F'),
                         #         }
                         )

    ds = EmaDataSet.load(emf, data_path, fmt='csv',
                         participant='file',
                         # path_groups=['Age', 'Gender'],
                         dtype={'CoSS': 'string'})
    print('ds= ', ds)
    test_path = work_path / 'test_save_nogroup'
    ds.save(test_path, fmt='csv', participant='TP', join_groups=True, group_join_str='_')
    print(f'Data set saved in {test_path}')

    # -------------------------------------TEST ONE group with ONE category
    emf = EmaFrame.setup(situations=sim_situations,
                         phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         groups={'Age': ['young'],  # *** ONE group, named
                                 # 'Gender': ('M', 'F'),
                                 }
                         )

    ds = EmaDataSet.load(emf, data_path, fmt='csv',
                         participant='file',
                         path_groups=['Age'],
                         dtype={'CoSS': 'string'})
    print('ds= ', ds)
    test_path = work_path / 'test_save_onegroup'
    ds.save(test_path, fmt='csv', participant='TP', join_groups=True, group_join_str='_')
    print(f'Data set saved in {test_path}')

    # -------------------------------------TEST ONE group with NO situation keys
    print('\n*** Test one group with no sitution keys')
    emf = EmaFrame.setup(#  situations=sim_situations,
                         #  phase_key='Phase',
                         attributes={'Speech': ['Very Hard',
                                                'Hard',
                                                'Easy',
                                                'Very Easy',
                                                'Perfect'],
                                     'Comfort': ['Bad',
                                                 'Not Good',
                                                 'Not Bad',
                                                 'Good']},
                         groups={'Age': ['young'],  # *** ONE group, named
                                 # 'Gender': ('M', 'F'),
                                 }
                         )

    ds = EmaDataSet.load(emf, data_path, fmt='csv',
                         participant='file',
                         path_groups=['Age'],
                         dtype={'CoSS': 'string'})
    print('ds= ', ds)
    test_path = work_path / 'test_save_no_Sit'
    ds.save(test_path, fmt='csv', participant='TP', join_groups=True, group_join_str='_')
    print(f'Data set saved in {test_path}')

    # -------------------------------------- TEST Empty attribute_grades
    print('\n*** Test one group, empty Attributes')
    emf = EmaFrame.setup(situations=sim_situations,
                         phase_key='Phase',
                         groups={'Age': ['young'],  # *** ONE group, named
                                 # 'Gender': ('M', 'F'),
                                 }
                         )
    ds = EmaDataSet.load(emf, data_path, fmt='csv',
                         participant='file',
                         path_groups=['Age'])
    print('ds= ', ds)
    test_path = work_path / 'test_save_no_Attr'
    ds.save(test_path, fmt='csv', participant='TP', join_groups=True, group_join_str='_')
    print(f'Data set saved in {test_path}')



"""This module estimates the effect measure Non-overlap of All Pairs (NAP),
to be used as a non-Bayesian measure of difference between Attribute Ratings.

The NAP measure is a non-parametric estimate of the probability that
Attribute Rating X < Y, given observed ordinal i.i.d. rating samples
x_n drawn from random variable X, and y_n drawn from Y,
X and Y represent two separate Situations.
The NAP measure is closely related to the Mann-Whitney statistic for ordinal data.

This module calculates point estimates and approximate confidence intervals for the NAP measure.

The confidence intervals are symmetric in probability.
For example, with a confidence level p=0.95,
the upper and lower tail probabilities are both = 0.025.
However, the central NAP point estimate is usually
NOT at the midpoint of the confidence interval.

There are many ways to calculate an approximate confidence interval for NAP.
Feng et al (2017) studied 29 different methods.
This module uses the "MW-N" variant, defined on page 2607 in their paper,
which showed good performance.
This variant was originally proposed by Newcombe (2009).

*** References:
R. I. Parker and K. Vannest.
An improved effect size for single-case research: Nonoverlap of all pairs.
Behavior Therapy, 40(4):357–367, 2009. doi: 10.1016/j.beth.2008.10.006

D. Feng, G. Cortese, and R. Baumgartner.
A comparison of confidence/credible interval methods for the area under the ROC curve
for continuous diagnostic tests with small sample size.
Statistical Methods in Medical Research, 26(6):2603–2621, 2017.
doi: 10.1177/0962280215602040

R. G. Newcombe.
Confidence intervals for an effect size measure based on the Mann–Whitney statistic.
Part 2: Asymptotic methods and evaluation.
Statistics in Medicine, 25:559–573, 2006. doi: 10.1002/sim.2324


*** Main module functions:
nap_pandas --- calculates point estimate and symmetric confidence intervals
    for Ordinal Grades stored as columns in a pandas DataFrame instance,
    grouped by values in other categorical columns, if desired.

nap_count --- calculates point estimate and symmetric confidence intervals
    for Ordinal Grades stored as arrays of grade Counts at each ordinal level.

nap_ci_low --- approximate lower limit of confidence interval
nap_ci_high --- approximate upper limit of confidence interval
    Can be used for either one-tailed or two-tailed confidence interval.

*** Version history:
* Version 0.9.4:
2022-11-06: Temp fix in nap_pandas for Pandas v 1.5.2 FutureWarning

* Version 0.9:
2022-03-24, calculate NAP from data stored in a pandas DataFrame

* Version 0.7:
2021-12-16, first functional version incl. confidence interval
2022-01-03, minor fix to give clearer warning message in case of nan result
"""
# *** publish this as package NAPcalc on PyPi separate from EmaCalc ? ********
import warnings  # *** needed? use logging?

import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.optimize import brentq
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST
# ham.logger.setLevel(logging.DEBUG)  # *** TEST


# ------------------------------------------------ Module exceptions
class NapError(RuntimeError):
    """Any type of exception while calculating NAP
    """


def nap_pandas(df, col, nap_cat=None,
               group_cols=None, grade_cols=None, p=0.95):
    """Calculate proportion of Non-overlapping Pairs = NAP result
    = estimate of P(X < Y), given observed ordinal i.i.d. rating samples
    x_n drawn from random variable X, and y_n drawn from Y.
    :param df: a Pandas.DataFrame object containing input data
    :param col: name of ONE column with categories for which NAP is to be calculated
    :param nap_cat: (optional) sequence of TWO categories (X, Y) in col.
        If None, df[col].dtype MUST be Categorical with exactly TWO categories.
    :param group_cols: (optional) sequence with name(s) of columns,
        for which results will be grouped.
    :param grade_cols: (optional) sequence with name(s) of columns with ordinal grades
        for which the NAP effect measure is calculated.
        If None, use ALL ordinal data columns in df
    :param p: (optional) confidence level for probability-symmetric confidence interval,
        i.e., upper == lower tail probability == (1-p) / 2.
    :return: a pandas DataFrame with all results stored in
        rows MultiIndex-ed by elements of group_cols, and
        columns MultiIndex-ed as (g_i, low), (g_i, nap), (g_i, high), where
            g_i is a grade label in grade_cols, and
            low is the lower conf.interval limit
            nap is the NAP point estimate,
            high is the upper conf.interval limit
        All calculated values show NAP for Second re. First category in col.
        Result table is sorted by categories in group_cols.
    """
    def calc_nap(group_name, group_df):
        """Calculate NAP results for given (sub-grouped) DataFrame
        :param group_name: single column or tuple of column names to mark in result
        :param group_df: a pd.DataFrame object
        :return: a dict with all NAP results as (key, value) pairs
        """
        # ***** check NaN grade columns *************
        if len(group_cols) == 1:  # *** temp fix for Pandas <= 1.5.2
            # not needed in future Pandas version !
            group_name = [group_name]  # because pd.groupby made it scalar in v. <= 1.5.2
        res = {g_col: g_cat
               for (g_col, g_cat) in zip(group_cols, group_name)}
        for a in grade_cols:
            x = group_df[group_df[col] == nap_cat[0]][a]
            y = group_df[group_df[col] == nap_cat[1]][a]
            x = x[x.notna()]
            y = y[y.notna()]
            n_x = len(x)
            n_y = len(y)
            if n_x == 0 or n_y == 0:
                res.update({(a, '[<'): np.nan,
                            (a, 'NAP'): np.nan,
                            (a, '<]'): np.nan})
            else:
                n_y_gt_x = sum(sum(y.gt(x_i) for x_i in x))
                n_y_eq_x = sum(sum(y.eq(x_i) for x_i in x))
                nap_point = (n_y_gt_x + n_y_eq_x / 2) / (n_x * n_y)
                z_tail = - norm.ppf((1. - p) / 2)
                res.update({(a, '[<'): nap_ci_low(nap_point, n_x, n_y, z_tail),
                            (a, 'NAP'): nap_point,
                            (a, '<]'): nap_ci_high(nap_point, n_x, n_y, z_tail)})
                # added three new columns for this attribute
        return res
        # ---------------------------------------------------------------------
    if nap_cat is None:
        if df[col].dtype.name == 'category' and len(df[col].dtype.categories) == 2:
            nap_cat = df[col].dtype.categories
        else:
            raise NapError(f'Cannot calculate NAP for column {col}. Must have exactly TWO categories')
    elif len(nap_cat) != 2:
        raise NapError(f'Cannot calculate NAP for nap_cat={nap_cat}. Must have exactly TWO categories')
    if grade_cols is None:
        grade_cols = [c for c in df.columns
                      if df[c].dtype == 'category'
                      and df[c].dtype.ordered]
    if len(grade_cols) == 0:
        raise NapError(f'No ordinal category columns found in given DataFrame.')
    if group_cols is None:
        group_cols = ()
        res = pd.DataFrame.from_records([calc_nap((), df)])
        # = DataFrame with ONE row of NAP results, no grouping
    elif len(group_cols) == 1:  # temp fix to avoid FutureWarning in Pandas <= 1.5.2
        # *** not needed when future df.groupby returns tuple (group_name,)
        res = pd.DataFrame.from_records([calc_nap(*df_g)
                                         for df_g in df.groupby(group_cols[0])])
        # = DataFrame with one row for each groupby category
        res.set_index(group_cols, inplace=True)
    else:
        res = pd.DataFrame.from_records([calc_nap(*df_g)
                                         for df_g in df.groupby(group_cols)])
        # = DataFrame with one row for each groupby category
        res.set_index(group_cols, inplace=True)
    col_ind = pd.MultiIndex.from_tuples(res.columns)
    return res.reindex(columns=col_ind)


# -----------------------------------------------------------------
def nap_count(x, y, p=0.95):  # *** not used in EmaCalc ***
    """Calculate proportion of Non-overlapping Pairs = NAP result,
    = estimate of P(X > Y), given observed ordinal i.i.d. rating samples
    x_n drawn from random variable X, and y_n drawn from Y,
    where X and Y represent two separate Situations.
    :param x: array of grade COUNTS for samples of ordinal random variable X
        x[i, ...] = number of observed samples of i-th ordinal grade
    :param y: array of COUNTS for ordinal random variable Y, similar
        y.shape == x.shape
    :param p: (optional) scalar confidence level
    :return: nap = array with
        nap[0, ...] = lower limit of symmetric confidence interval
        nap[1, ...] = point estimate of P(X > Y)
        nap[2, ...] = upper limit of symmetric confidence interval
        nap.shape[1:] == x.shape[1:] == y.shape[1:]
    """
    result_shape = x.shape[1:]
    cum_y = np.cumsum(y, axis=0)
    # cum_y[i] = n Y <= i
    n_y = cum_y[-1]
    n_x = np.sum(x, axis=0)
    n_x_gt_y = np.sum(x[1:, ...] * cum_y[:-1, ...], axis=0)
    # = sum_i number of pairs where X == i AND Y < i
    n_x_eq_y = np.sum(x * y, axis=0)
    # = n pairs where X == Y
    n_pairs = n_x * n_y
    if np.any(n_pairs == 0):
        logger.warning('NO rating pairs -> undefined NAP value in some case(s).')
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')  # suppress standard warning for div by zero
        nap_point = (n_x_gt_y + 0.5 * n_x_eq_y) / n_pairs
    z_tail = norm.isf((1. - p) / 2)  # upper tail quantile
    ci_low = np.array([nap_ci_low(p, nx, ny, z_tail)
                       for (p, nx, ny) in zip(nap_point.reshape((-1,)),
                                              n_x.reshape((-1,)),
                                              n_y.reshape((-1,)))])
    ci_high = np.array([nap_ci_high(p, nx, ny, z_tail)
                       for (p, nx, ny) in zip(nap_point.reshape((-1,)),
                                              n_x.reshape((-1,)),
                                              n_y.reshape((-1,)))])
    return np.array([ci_low.reshape(result_shape),
                     nap_point,
                     ci_high.reshape(result_shape)])


def nap_ci_low(a_hat, nx, ny, z_quantile):
    """Calculate lower limit of symmetric confidence interval for NAP
    :param a_hat: scalar point estimate of P(X > Y)
    :param nx: number of X observations
    :param ny: number of Y observations
    :param z_quantile: quantile at higher tail of standard-normal distribution
    :return: scalar ci_low

    Method: MW-N as described by Feng et al (2017),
        using the improved variant solving their nonlinear Eq. (3)
    """
    def fun_3(a):
        """Function to solve Feng et al Eq (3)
        for the LOWER confidence limit, where 0 <= a < a_hat
        :param a: scalar solution candidate
        :return: scalar function value
        """
        return a_hat - a - z_quantile * nap_stdev(a, nx, ny)
    # --------------------------------------------------
    if nx == 0 or ny == 0:
        return np.nan
    if np.isclose(a_hat, 0.):
        return 0.
    ci_lim, res = brentq(fun_3, 0., a_hat, full_output=True)
    if res.converged:
        return ci_lim
    else:
        logger.warning('Confidence-interval calculation did not converge')
        return np.nan


def nap_ci_high(a_hat, nx, ny, z_quantile):
    """Calculate upper limit of symmetric confidence interval for NAP
    :param a_hat: scalar point estimate of P(X > Y)
    :param nx: number of X observations
    :param ny: number of Y observations
    :param z_quantile: quantile at higher tail of standard-normal distribution
    :return: scalar ci_low

    Method: MW-N as described by Feng et al (2017),
        using the improved variant solving their nonlinear Eq. (3)
    """
    def fun_3(a):
        """Function to solve Feng et al Eq (3)
        for the UPPER confidence limit, where a_hat < a <= 1.
        :param a: scalar solution candidate
        :return: scalar function value
        """
        return a - a_hat - z_quantile * nap_stdev(a, nx, ny)
    # --------------------------------------------------
    if nx == 0 or ny == 0:
        return np.nan
    if np.isclose(a_hat, 1.):
        return 1.
    ci_lim, res = brentq(fun_3, a_hat, 1., full_output=True)
    if res.converged:
        return ci_lim
    else:
        logger.warning('Confidence-interval calculation did not converge')
        return np.nan


def nap_stdev(a, nx, ny):
    """Help function to estimate st.dev. of NAP value
    :param a: scalar candidate value for NAP
    :param nx: number of X observations
    :param ny: number of Y observations
    :return: s = sqrt(var(A)), estimated at point NAP A == a

    Method: var_H defined just before Eq (3) in Feng et al (2017)
    """
    n_star = (nx + ny) / 2 - 1
    # = symmetrized as suggested by Newcombe (2009)
    v = a * (1. - a)
    v *= 1. + n_star * ((1. - a) / (2. - a) + a / (1. + a))
    v /= nx * ny
    return np.sqrt(v)


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    print('*** Testing nap_pandas ***')
    df = pd.DataFrame({'HA': list('ABABAA'),
                       }, dtype='category')  # not ordered
    df['Speech'] = pd.Series(['Bad', 'Good', 'Good', 'Good', 'Bad', 'Bad'],
                             dtype=pd.CategoricalDtype(categories=['Bad', 'Good'],
                                                       ordered=True))
    df['Comfort'] = pd.Series(['Bad', 'OK', 'Bad', 'Good', 'Good', 'OK'],
                             dtype=pd.CategoricalDtype(categories=['Bad', 'OK', 'Good'],
                                                       ordered=True))
    nap = nap_pandas(df, 'HA')
    print(nap)
    print('')
    df['Group'] = 'All'
    nap = nap_pandas(df, 'HA', group_cols=['Group'])
    print(nap)
    print('')
    nap = nap_pandas(df, 'HA', nap_cat=('B', 'A'), group_cols=['Group'])
    print(nap)
    print('')
    df['Speech'][0] = np.nan
    nap = nap_pandas(df, 'HA', nap_cat=('B', 'A'), group_cols=['Group'])
    print(nap)
    print('')

    print('*** Testing nap_statistic ***')
    x_count = np.array([1, 2, 3, 4, 5])
    y_count = np.array([3, 3, 3, 3, 3])
    print(f'NAP result =\n\t{nap_count(x_count, y_count)}')

    print('*** Testing nap_statistic ***')
    x_count = np.array([0, 0, 0, 5])
    y_count = np.array([4, 3, 2, 0])
    print(f'NAP result =\n\t{nap_count(x_count, y_count)}')


